import os
import stat
import errno
import time
from .ops import Operations, AsyncOperations, FuseOSError, logger


class FUSEr(Operations):
    def __init__(self, fs, path, ready_file=False, **kwargs):
        self.fs = fs
        self.cache = {}
        self.root = path.rstrip("/") + "/"
        self.counter = 0
        logger.info("Starting FUSE at %s", path)
        self._ready_file = ready_file

    def getattr(self, path, fh=None):
        logger.debug("getattr %s", path)
        if self._ready_file and path in ["/.fuse_ready", ".fuse_ready", "/.fuze_mounted", ".fuze_mounted"]: return {"type": "file", "st_size": 5}
        path = "".join([self.root, path.lstrip("/")]).rstrip("/")
        try: info = self.fs.info(path)
        except FileNotFoundError: raise FuseOSError(errno.ENOENT)
        data = {"st_uid": info.get("uid", 1000), "st_gid": info.get("gid", 1000)}
        perm = info.get("mode", 0o777)

        if info["type"] != "file":
            data["st_mode"] = stat.S_IFDIR | perm
            data["st_size"] = 0
            data["st_blksize"] = 0
        else:
            data["st_mode"] = stat.S_IFREG | perm
            data["st_size"] = info["size"]
            data["st_blksize"] = 5 * 2 ** 20
            data["st_nlink"] = 1
        data["st_atime"] = time.time()
        data["st_ctime"] = time.time()
        data["st_mtime"] = time.time()
        return data

    def readdir(self, path, fh):
        logger.debug("readdir %s", path)
        path = "".join([self.root, path.lstrip("/")])
        files = self.fs.ls(path, False)
        files = [os.path.basename(f.rstrip("/")) for f in files]
        return [".", ".."] + files

    def mkdir(self, path, mode):
        path = "".join([self.root, path.lstrip("/")])
        self.fs.mkdir(path)
        return 0

    def rmdir(self, path):
        path = "".join([self.root, path.lstrip("/")])
        self.fs.rmdir(path)
        return 0

    def read(self, path, size, offset, fh):
        logger.debug("read %s", (path, size, offset))
        if self._ready_file and path in ["/.fuse_ready", ".fuse_ready", "/.fuze_mounted", ".fuze_mounted"]: return b"ready"
        f = self.cache[fh]
        f.seek(offset)
        return f.read(size)

    def write(self, path, data, offset, fh):
        logger.debug("write %s", (path, offset))
        f = self.cache[fh]
        f.write(data)
        return len(data)

    def create(self, path, flags, fi=None):
        logger.debug("create %s", (path, flags))
        fn = "".join([self.root, path.lstrip("/")])
        self.fs.touch(fn)  # OS will want to get attributes immediately
        f = self.fs.open(fn, "wb")
        self.cache[self.counter] = f
        self.counter += 1
        return self.counter - 1

    def open(self, path, flags):
        logger.debug("open %s", (path, flags))
        fn = "".join([self.root, path.lstrip("/")])
        if flags % 2 == 0: mode = "rb"
        else: mode = "wb"
        self.cache[self.counter] = self.fs.open(fn, mode)
        self.counter += 1
        return self.counter - 1

    def truncate(self, path, length, fh=None):
        fn = "".join([self.root, path.lstrip("/")])
        if length != 0: raise NotImplementedError
        # maybe should be no-op since open with write sets size to zero anyway
        self.fs.touch(fn)

    def unlink(self, path):
        fn = "".join([self.root, path.lstrip("/")])
        try: self.fs.rm(fn, False)
        except (IOError, FileNotFoundError): raise FuseOSError(errno.EIO)

    def release(self, path, fh):
        try:
            if fh in self.cache:
                f = self.cache[fh]
                f.close()
                self.cache.pop(fh)
        except Exception as e:
            print(e)
        return 0

    def chmod(self, path, mode):
        if hasattr(self.fs, "chmod"):
            path = "".join([self.root, path.lstrip("/")])
            return self.fs.chmod(path, mode)
        raise NotImplementedError

class FUSErCache(FUSEr):
    """
    Uses lazy.io.cachez
    """
    def __init__(self, fs, path, cache_dir: str, ready_file: bool = False):
        from lazy.io.cachez import Cache
        self.fs = fs
        self.cache = Cache(directory=cache_dir)
        self.root = path.rstrip("/") + "/"
        self.counter = 0
        logger.info("Starting FUSECache at %s", path)
        self._ready_file = ready_file



class AsyncFUSEr(AsyncOperations):
    def __init__(self, fs, path, ready_file=False, **kwargs):
        self.fs = fs
        self.cache = {}
        self.root = path.rstrip("/") + "/"
        self.counter = 0
        logger.info("Starting FUSE at %s", path)
        self._ready_file = ready_file

    async def getattr(self, path, fh=None):
        logger.debug("getattr %s", path)
        if self._ready_file and path in ["/.fuse_ready", ".fuse_ready", "/.fuze_mounted", ".fuze_mounted"]: return {"type": "file", "st_size": 5}
        path = "".join([self.root, path.lstrip("/")]).rstrip("/")
        try: info = await self.fs._info(path)
        except FileNotFoundError: raise FuseOSError(errno.ENOENT)
        data = {"st_uid": info.get("uid", 1000), "st_gid": info.get("gid", 1000)}
        perm = info.get("mode", 0o777)

        if info["type"] != "file":
            data["st_mode"] = stat.S_IFDIR | perm
            data["st_size"] = 0
            data["st_blksize"] = 0
        else:
            data["st_mode"] = stat.S_IFREG | perm
            data["st_size"] = info["size"]
            data["st_blksize"] = 5 * 2 ** 20
            data["st_nlink"] = 1
        data["st_atime"] = time.time()
        data["st_ctime"] = time.time()
        data["st_mtime"] = time.time()
        return data

    async def readdir(self, path, fh):
        logger.debug("readdir %s", path)
        path = "".join([self.root, path.lstrip("/")])
        files = await self.fs._ls(path, False)
        files = [os.path.basename(f.rstrip("/")) for f in files]
        return [".", ".."] + files

    async def mkdir(self, path, mode):
        path = "".join([self.root, path.lstrip("/")])
        await self.fs._mkdir(path)
        return 0

    async def rmdir(self, path):
        path = "".join([self.root, path.lstrip("/")])
        await self.fs._rmdir(path)
        return 0

    async def read(self, path, size, offset, fh):
        logger.debug("read %s", (path, size, offset))
        if self._ready_file and path in ["/.fuse_ready", ".fuse_ready", "/.fuze_mounted", ".fuze_mounted"]: return b"ready"
        f = self.cache[fh]
        f.seek(offset)
        return f.read(size)

    async def write(self, path, data, offset, fh):
        logger.debug("write %s", (path, offset))
        f = self.cache[fh]
        f.write(data)
        return len(data)

    async def create(self, path, flags, fi=None):
        logger.debug("create %s", (path, flags))
        fn = "".join([self.root, path.lstrip("/")])
        self.fs.touch(fn)  # OS will want to get attributes immediately
        f = self.fs.open(fn, "wb")
        self.cache[self.counter] = f
        self.counter += 1
        return self.counter - 1

    async def open(self, path, flags):
        logger.debug("open %s", (path, flags))
        fn = "".join([self.root, path.lstrip("/")])
        if flags % 2 == 0: mode = "rb"
        else: mode = "wb"
        self.cache[self.counter] = self.fs.open(fn, mode)
        self.counter += 1
        return self.counter - 1

    async def truncate(self, path, length, fh=None):
        fn = "".join([self.root, path.lstrip("/")])
        if length != 0: raise NotImplementedError
        # maybe should be no-op since open with write sets size to zero anyway
        self.fs.touch(fn)

    async def unlink(self, path):
        fn = "".join([self.root, path.lstrip("/")])
        try: await self.fs._rm(fn, False)
        except (IOError, FileNotFoundError): raise FuseOSError(errno.EIO)

    async def release(self, path, fh):
        try:
            if fh in self.cache:
                f = self.cache[fh]
                f.close()
                self.cache.pop(fh)
        except Exception as e:
            print(e)
        return 0

    async def chmod(self, path, mode):
        if hasattr(self.fs, "chmod"):
            path = "".join([self.root, path.lstrip("/")])
            return await self.fs._chmod(path, mode)
        raise NotImplementedError


class AsyncFUSErCache(AsyncFUSEr):
    """
    Uses lazy.io.cachez
    """
    def __init__(self, fs, path, cache_dir: str, ready_file: bool = False):
        from lazy.io.cachez import Cache
        self.fs = fs
        self.cache = Cache(directory=cache_dir)
        self.root = path.rstrip("/") + "/"
        self.counter = 0
        logger.info("Starting AsyncFUSECache at %s", path)
        self._ready_file = ready_file



import io
import inspect
import reprlib
import pathlib
import types

from copy import copy
from glob import glob
from subprocess import PIPE, Popen, TimeoutExpired
from lazy.types import classproperty
from lazy.serialize import OrJson
from lazy.io import pathz, get_path, PathLike, PathzPath





"""
Handler to check Class Imports
"""
import os
import sys
import importlib
import subprocess
import pkg_resources


def get_variable_separator():
    """
    Returns the environment variable separator for the current platform.
    :return: Environment variable separator
    """
    if sys.platform.startswith('win'): return ';'
    return ':'



class LazyLibType(type):
    
    def get_requirement(cls, name: str, clean: bool = True) -> str:
        if not clean: return name.strip()
        return name.split('=')[0].replace('>','').replace('<','').strip()
    
    def install_library(cls, library: str, upgrade: bool = True):
        pip_exec = [sys.executable, '-m', 'pip', 'install']
        if '=' not in library or upgrade: pip_exec.append('--upgrade')
        pip_exec.append(library)
        return subprocess.check_call(pip_exec, stdout=subprocess.DEVNULL)
        
    def is_available(cls, library: str) -> bool:
        """ Checks whether a Python Library is available."""
        try:
            _ = pkg_resources.get_distribution(library)
            return True
        except pkg_resources.DistributionNotFound: return False
    
    def is_imported(cls, library: str) -> bool:
        """ Checks whether a Python Library is currently imported."""
        return library in sys.modules
    
    def _ensure_lib_imported(cls, library: str):
        clean_lib = cls.get_requirement(library, True)
        if not cls.is_imported(clean_lib): sys.modules[clean_lib] = importlib.import_module(clean_lib)
        #if clean_lib not in sys.modules: sys.modules[clean_lib] = importlib.import_module(clean_lib)    
        return sys.modules[clean_lib]
        
    
    def import_lib(cls, library: str, resolve_missing: bool = True, require: bool = False, upgrade: bool = False):
        """ Lazily resolves libs.
            if available, returns the sys.modules[library]
            if missing and resolve_missing = True, will lazily install
        else:
            if require: raise ImportError
            returns None
        """
        clean_lib = cls.get_requirement(library, True)
        if not cls.is_available(clean_lib):
            if require and not resolve_missing: raise ImportError(f"Required Lib {library} is not available.")
            if not resolve_missing: return None
            cls.install_library(library, upgrade=upgrade)
        return cls._ensure_lib_imported(library)
    
    def get_binary_path(cls, executable: str):
        """
        Searches for a binary named `executable` in the current PATH. If an executable is found, its absolute path is returned
        else None.
        :param executable: Name of the binary
        :return: Absolute path or None
        """
        if 'PATH' not in os.environ: return None
        for directory in os.environ['PATH'].split(get_variable_separator()):
            binary = os.path.abspath(os.path.join(directory, executable))
            if os.path.isfile(binary) and os.access(binary, os.X_OK): return binary
        return None
    
    def is_exec_available(cls, executable: str) -> bool:
        return bool(cls.get_binary_path(executable) is not None)

    def __getattr__(cls, key):
        """
            LazyLib.is_avail_tensorflow -> bool
            LazyLib.tensorflow -> sys.modules[tensorflow] or None
            
            LazyLib.is_avail_pydantic -> True (since its installed with this lib)
            LazyLib.is_imported_tensorflow 
            LazyLib.is_exec_avail_bash -> True
        """
        if key.startswith('is_avail_'):
            lib_name = key.split('is_avail_')[-1].strip()
            return cls.is_available(lib_name)
        
        if key.startswith('is_imported_'):
            lib_name = key.split('is_imported_')[-1].strip()
            return cls.is_imported(lib_name)
        
        if key.startswith('is_exec_avail_'):
            exec_name = key.split('is_exec_avail_')[-1].strip()
            return cls.is_exec_available(exec_name)
        
        return cls.import_lib(key, resolve_missing=False, require=False)
        
class LazyLib(metaclass=LazyLibType):
    pass


import os
import sys
import subprocess
import logging
import platform
from pathlib import Path
from typing import Union, List, Type

from logz import get_logger

logger = get_logger('procs')


"""
Deprecated Exec Functions
"""
def exec_cmd(cmd, raise_error: bool = True):
    try:
        out = subprocess.check_output(cmd, shell=True)
        if isinstance(out, bytes): out = out.decode('utf8')
        return out.strip()
    except Exception as e:
        if not raise_error: return ""
        raise e

def exec_daemon(cmd: Union[List[str], str], stdout = subprocess.PIPE, stderr = subprocess.STDOUT, set_proc_uid: bool = False, *args, **kwargs):
    if isinstance(cmd, str): cmd = [cmd]
    if set_proc_uid and platform.system() != 'Darwin': return subprocess.Popen(cmd, stdout = stdout, stderr = stderr, preexec_fn = lambda: os.setuid(1), *args, **kwargs)
    return subprocess.Popen(cmd, stdout = stdout, stderr = stderr, *args, **kwargs)

def exec_run(cmd: Union[List[str], str], *args, **kwargs):
    if isinstance(cmd, str): cmd = [cmd]
    return subprocess.run(cmd, *args, **kwargs)

def exec_out(cmd: Union[List[str], str], shell: bool = True, *args, **kwargs):
    if isinstance(cmd, str): cmd = [cmd]
    return subprocess.check_output(cmd, shell = shell, *args, **kwargs)


def exec_shell(cmd): return os.system(cmd)


"""
Handler to check Class Imports
"""
import os
import sys
import importlib
import subprocess
import pkg_resources
import pathlib

from typing import List, Type, Any, Union, Dict
from types import ModuleType

_root_path = pathlib.Path(__file__).parent
_scriptz_path = _root_path.joinpath('scriptz')
_bash_scriptz_path = _scriptz_path.joinpath('bash')


def get_variable_separator():
    """
    Returns the environment variable separator for the current platform.
    :return: Environment variable separator
    """
    if sys.platform.startswith('win'): return ';'
    return ':'

class PkgInstall:
    win: str = 'choco install [flags]'
    mac: str = 'brew [flags] install '
    linux: str = 'apt-get -y [flags] install'

    @classmethod
    def get_args(cls, binary: str, flags: List[str] = None):
        """
        Builds the install args for the system
        """
        flag_str = ' '.join(flags) if flags else '' 
        if sys.platform.startswith('win'):
            b = cls.win + ' ' + binary
            b = b.replace('[flags]', flag_str)
        
        if sys.platform.startswith('linux'):
            b = cls.linux + ' ' + binary
            b = b.replace('[flags]', flag_str)

        if sys.platform.startswith('darwin'):
            b = cls.mac + ' ' + binary
            b = b.replace('[flags]', flag_str)
        return [i.strip() for i in b.split(' ') if i.strip()]
        


class LibType(type):
    
    def get_requirement(cls, name: str, clean: bool = True) -> str:
        # Replaces '-' with '_'
        # for any library such as tensorflow-text -> tensorflow_text
        name = name.replace('-', '_')
        if not clean: return name.strip()
        return name.split('=')[0].replace('>','').replace('<','').strip()
    
    def install_library(cls, library: str, upgrade: bool = True):
        pip_exec = [sys.executable, '-m', 'pip', 'install']
        if '=' not in library or upgrade: pip_exec.append('--upgrade')
        pip_exec.append(library)
        return subprocess.check_call(pip_exec, stdout=subprocess.DEVNULL)
    
    def install_binary(cls, binary: str, flags: List[str] = None):
        if cls.get_binary_path(binary): return
        args = PkgInstall.get_args(binary, flags)
        return subprocess.check_call(args, stdout=subprocess.DEVNULL)
        
    def is_available(cls, library: str) -> bool:
        """ Checks whether a Python Library is available."""
        try:
            _ = pkg_resources.get_distribution(library)
            return True
        except pkg_resources.DistributionNotFound: return False
    
    def is_imported(cls, library: str) -> bool:
        """ Checks whether a Python Library is currently imported."""
        return library in sys.modules
    
    def _ensure_lib_imported(cls, library: str):
        clean_lib = cls.get_requirement(library, True)
        if not cls.is_imported(clean_lib): sys.modules[clean_lib] = importlib.import_module(clean_lib)
        return sys.modules[clean_lib]
    
    def _ensure_lib_installed(cls, library: str, pip_name: str = None, upgrade: bool = False):
        clean_lib = cls.get_requirement(library, True)
        if not cls.is_available(clean_lib):
            cls.install_library(pip_name or library, upgrade=upgrade)

    def _ensure_binary_installed(cls, binary: str, flags: List[str] = None):
        return cls.install_binary(binary, flags)
    
    def import_lib(cls, library: str, pip_name: str = None, resolve_missing: bool = True, require: bool = False, upgrade: bool = False) -> ModuleType:
        """ Lazily resolves libs.

            if pip_name is provided, will install using pip_name, otherwise will use libraryname

            ie ->   Lib.import_lib('fuse', 'fusepy') # if fusepy is not expected to be available, and fusepy is the pip_name
                    Lib.import_lib('fuse') # if fusepy is expected to be available
            
            returns `fuse` as if you ran `import fuse`
        
            if available, returns the sys.modules[library]
            if missing and resolve_missing = True, will lazily install
        else:
            if require: raise ImportError
            returns None
        """
        clean_lib = cls.get_requirement(library, True)
        if not cls.is_available(clean_lib):
            if require and not resolve_missing: raise ImportError(f"Required Lib {library} is not available.")
            if not resolve_missing: return None
            cls.install_library(pip_name or library, upgrade=upgrade)
        return cls._ensure_lib_imported(library)
    
    def import_module(cls, name: str, library: str = None, pip_name: str = None, resolve_missing: bool = True, require: bool = False, upgrade: bool = False) -> ModuleType:
        """ Lazily resolves libs and imports the name, aliasing
            immportlib.import_module

            ie ->   Lib.import_module('tensorflow.io.gfile', 'tensorflow') # if tensorflow is not expected to be available
                    Lib.import_module('tensorflow.io.gfile') # if tensorflow is expected to be available
            returns tensorflow.io.gfile
        """
        if library:
            cls.import_lib(library, pip_name, resolve_missing, require, upgrade)
            return importlib.import_module(name, package=library)
        return importlib.import_module(name)

    def import_module_attr(cls, name: str, module_name: str, library: str = None, pip_name: str = None, resolve_missing: bool = True, require: bool = False, upgrade: bool = False) -> Any:
        """ Lazily resolves libs and imports the name, aliasing
            immportlib.import_module
            Returns an attribute from the module

            ie ->   Lib.import_module_attr('GFile', 'tensorflow.io.gfile', 'tensorflow') # if tensorflow is not expected to be available
                    Lib.import_module_attr('GFile', 'tensorflow.io.gfile') # if tensorflow is expected to be available
            returns GFile
        """
        mod = cls.import_module(name=module_name, library=library, pip_name=pip_name, resolve_missing = resolve_missing, require = require, upgrade = upgrade)
        return getattr(mod, name)
    
    def import_cmd(cls, binary: str, resolve_missing: bool = True, require: bool = False, flags: List[str] = None):
        """ Lazily builds a lazy.Cmd based on binary
        
            if available, returns the lazy.Cmd(binary)
            if missing and resolve_missing = True, will lazily install in host system
        else:
            if require: raise ImportError
            returns None
        """
        if not cls.is_exec_available(binary):
            if require and not resolve_missing: raise ImportError(f"Required Executable {binary} is not available.")
            if not resolve_missing: return None
            cls.install_binary(binary, flags=flags)
        from lazy.cmd import Cmd
        return Cmd(binary=binary)

    def get_binary_path(cls, executable: str):
        """
        Searches for a binary named `executable` in the current PATH. If an executable is found, its absolute path is returned
        else None.
        :param executable: Name of the binary
        :return: Absolute path or None
        """
        if 'PATH' not in os.environ: return None
        for directory in os.environ['PATH'].split(get_variable_separator()):
            binary = os.path.abspath(os.path.join(directory, executable))
            if os.path.isfile(binary) and os.access(binary, os.X_OK): return binary
        return None
    
    def is_exec_available(cls, executable: str) -> bool:
        return bool(cls.get_binary_path(executable) is not None)
    
    def run_bash(cls, cmd: str = None, bash_name: str = 'bash', *args, **kwargs):
        """
        Utility to run bash command

        Lib.run_bash('/path/to/script.sh') -> bash /path/to/script.sh
        """

        from lazy.cmd import Cmd
        bash = Cmd(bash_name)
        return bash(cmd, *args, **kwargs).val
    
    def run_bash_script(cls, script_name: str, *args, **kwargs):
        if not script_name.endswith('.sh'): script_name += '.sh'
        scriptz = _bash_scriptz_path.joinpath(script_name)
        assert scriptz.exists(), f'Invalid Script: {script_name} does not exist.'
        cls.run_bash(cmd=scriptz.as_posix())
    
    def run_fusemount(cls, bucket: str, mountpoint: str, auth_config: Dict[Any, Any] = None, *cliargs, **kwargs):
        """
        Runs fuse_v1
        Base Args:
        BUCKET=$1
        MOUNT_PATH=$2

        For S3:
        S3_TYPE=$3
        """

        from lazy.configz import CloudAuthz
        from lazy.cmd import Cmd

        uris = bucket.split('://', maxsplit=1)
        provider, bucket_path = uris[0], uris[-1]
        if auth_config is not None: CloudAuthz.update_authz(**auth_config)
        bash = Cmd('bash')
        pathlib.Path(mountpoint).mkdir(exist_ok=True, parents=True)
        CloudAuthz.set_authz_env()
        if provider == 'gs':
            scriptz = _bash_scriptz_path.joinpath('gcsfuse.sh')
            return bash(scriptz.as_posix(), bucket_path, mountpoint, *cliargs, **kwargs).val
        
        scriptz = _bash_scriptz_path.joinpath('s3fuse.sh')
        if provider == 's3': return bash(scriptz.as_posix(), bucket_path, mountpoint, 'aws', *cliargs, **kwargs).val
        return bash(scriptz.as_posix(), bucket_path, mountpoint, provider, *cliargs, **kwargs).val





    def __getattr__(cls, key):
        """
            Lib.is_avail_tensorflow -> bool
            Lib.tensorflow -> sys.modules[tensorflow] or None
            
            Lib.is_avail_pydantic -> True (since its installed with this lib)
            Lib.is_imported_tensorflow 
            Lib.is_avail_bin_bash -> True
            Lib.is_avail_exec_bash -> True
        """
        if key.startswith('is_avail_bin_'):
            exec_name = key.split('is_avail_bin_')[-1].strip()
            return cls.is_exec_available(exec_name)
        
        if key.startswith('is_avail_exec_'):
            exec_name = key.split('is_avail_exec_')[-1].strip()
            return cls.is_exec_available(exec_name)
        
        if key.startswith('is_avail_lib_'):
            lib_name = key.split('is_avail_lib_')[-1].strip()
            return cls.is_available(lib_name)
        
        if key.startswith('is_avail_'):
            lib_name = key.split('is_avail_')[-1].strip()
            return cls.is_available(lib_name)
        
        if key.startswith('is_imported_'):
            lib_name = key.split('is_imported_')[-1].strip()
            return cls.is_imported(lib_name)
        
        if key.startswith('cmd_'):
            binary_name = key.split('cmd_')[-1].strip()
            return cls.import_cmd(binary=binary_name)
        
        return cls.import_lib(key, resolve_missing=False, require=False)
    
    def get(cls, name: str, attr_name: str = None, pip_name: str = None, resolve_missing: bool = True) -> Union[ModuleType, Any]:
        """
        Resolves the import based on params.

        Importing a module:
        Lib.get('tensorflow') -> tensorflow Module
        Lib.get('fuse', pip_name='fusepy') -> fuse Module
        
        Importing a submodule:
        Lib.get('tensorflow.io') -> io submodule

        Importing something from within a submodule:
        Lib.get('tensorflow.io.gfile', 'GFile') -> GFile class
        """
        if attr_name: 
            lib_name = name.split('.', 1)[0]
            return cls.import_module_attr(attr_name, module_name = name, library = lib_name, pip_name = pip_name, resolve_missing = resolve_missing)
        if '.' in name: 
            lib_name = name.split('.', 1)[0]
            return cls.import_module(name, library=lib_name, pip_name=pip_name, resolve_missing=resolve_missing)
        return cls.import_lib(name, pip_name=pip_name, resolve_missing=resolve_missing)
    
    def _parse_name(cls, name: str) -> Dict[str, str]:
        """
        Resolves the name into a dict = 
        {
            'library': str,
            'pip_name': Optional[str],
            'module_name': Optional[str],
            'attr_name': Optional[str]
        }
        
        - module.submodule:attribute
        - pip_name|module.submodule:attribute # if pip_name is not the same
        
        Lib['tensorflow']                   -> {'library': 'tensorflow'}
        Lib['fusepy|fuse']                  -> {'library': 'fuse', 'pip_name': 'fusepy'}
        Lib['tensorflow.io']                -> {'library': 'tensorflow', 'module_name': 'tensorflow.io'}
        Lib['tensorflow.io.gfile:GFile']    -> {'library': 'tensorflow', 'module_name': 'tensorflow.io.gfile', 'attr_name': GFile}
        """
        rez = {'library': '', 'pip_name': None, 'module_name': None, 'attr_name': None}
        _name = name.strip()
        if ':' in _name:
            _s = _name.split(':', 1)
            rez['attr_name'] = _s[-1]
            _name = _s[0]
        if '|' in _name:
            _s = _name.split('|', 1)
            rez['pip_name'] = _s[0]
            _name = _s[-1]
        if '.' in _name:
            _s = _name.split('.', 1)
            rez['module_name'] = _name
            rez['library'] = _s[0]
        else: 
            #rez['module_name'] = _name
            rez['library'] = _name
        return rez
    
    def __getitem__(cls, name: str) -> ModuleType:
        """
        Resolves the import based on params. 
        Will automatically assume as resolve_missing = True, require = True
        
        The naming scheme is resolved as
        
        - module.submodule:attribute
        - pip_name|module.submodule:attribute # if pip_name is not the same

        Importing a module:
        Lib['tensorflow'] -> tensorflow Module
        Lib['fusepy|fuse'] -> fuse Module
        
        Importing a submodule:
        Lib['tensorflow.io'] -> io submodule

        Importing something from within a submodule:
        Lib['tensorflow.io.gfile:GFile'] -> GFile class
        """
        r = cls._parse_name(name)
        if r.get('attr_name'): return cls.import_module_attr(r['attr_name'], module_name = r['module_name'], library = r['library'], pip_name = r['pip_name'], resolve_missing = True, require = True)
        if r.get('module_name'): return cls.import_module(name = r['module_name'], library = r['library'], pip_name = r['pip_name'], resolve_missing = True, require = True)
        return cls.import_lib(r['library'], pip_name = r['pip_name'], resolve_missing = True, require = True)



        
class Lib(metaclass=LibType):
    pass



READ = 'r'
WRITE = 'w'
APPEND = 'a'

READ_BINARY = 'rb'
WRITE_BINARY = 'wb'
BINARY_MODES = (READ_BINARY, WRITE_BINARY)

NEWLINE = '\n'
BINARY_NEWLINE = b'\n'

WHENCE_START = 0
WHENCE_CURRENT = 1
WHENCE_END = 2
WHENCE_CHOICES = (WHENCE_START, WHENCE_CURRENT, WHENCE_END)

class Mode:
    read = READ
    read_binary = READ_BINARY
    write = WRITE
    write_binary = WRITE_BINARY
    append = APPEND
    binary_modes = BINARY_MODES


    


"""
Cloud Path extensions if libs are available
"""

from lazycls.ext._imports import LazyLib
from lazycls.ext.pathio.base import Path
from lazycls.ext.pathio.base import PathLike

if LazyLib.is_avail_smartopen:
    pass

if LazyLib.is_avail_tensorflow:
    pass




from . import types
from . import prop
from . import utils
from . import serializers
from . import models
from . import base
from . import funcs
from . import envs
from . import timez
from . import io
from . import ext
from . import typec

from .prop import classproperty, ClasspropertyMeta
from .utils import toPath, to_path
from .models import (
    BaseCls,
    BaseModel,
    BaseLazy,
    Field
)
from .timez import TimeCls

from .base import (
    LazyCls,
    create_lazycls,
    clear_lazy_models,
    set_modulename
)
from .funcs import (
    caseCamelToSnake,
    camelcase_to_snakecase,
)

from .envs import (
    loadEnvFile, 
    load_env_file,
    toEnv,
    to_env,
    Env
)

from .io import Path

__all__ = [
    'classproperty',
    'ClasspropertyMeta',
    'BaseCls',
    'BaseModel',
    'BaseLazy',
    'Field',
    'LazyCls',
    'create_lazycls',
    'Env',
    'Path',

]


__all__ = (
    'Serializer',
    'SerializerB',
    'Mode'
)

from typing import Any
from logz import get_logger
from .static import Mode, Defaults

logger = get_logger('serialize')

class Serializer:
    encoding: str = 'utf-8'
    binary: bool = False
    read_mode: str = Mode.read
    write_mode: str = Mode.write
    append_mode: str = Mode.append
    default_mode: str = Mode.read
    default_value: Any = None    
    cloud_supported: bool = False
    async_supported: bool = False
    
    @classmethod
    def _validate_inputs(cls, obj, *args, default: Any = None, **kwargs):
        raise NotImplementedError
    
    @classmethod
    def _validate_outputs(cls, value: Any, *args, **kwargs):
        raise NotImplementedError
    
    @classmethod
    def _encode(cls, obj, *args, default: Any = None, **kwargs):
        raise NotImplementedError
    
    @classmethod
    def _decode(cls, obj, *args, **kwargs):
        raise NotImplementedError
    
    @classmethod
    async def _async_encode(cls, obj, *args, default: Any = None, **kwargs):
        raise NotImplementedError
    
    @classmethod
    async def _async_decode(cls, obj, *args, **kwargs):
        raise NotImplementedError
    
    @classmethod
    def dumps(cls, obj, *args, default: Any = None, **kwargs):
        return cls._encode(obj, *args, default = default, **kwargs)

    @classmethod
    def loads(cls, obj, *args, **kwargs):
        return cls._decode(obj, *args, **kwargs)
    
    @classmethod
    async def async_dumps(cls, obj, *args, default: Any = None, **kwargs):
        if not cls.async_supported: raise Exception
        return await cls._async_encode(obj, *args, default = default, **kwargs)

    @classmethod
    async def async_loads(cls, obj, *args, **kwargs):
        if not cls.async_supported: raise Exception
        return await cls._async_decode(obj, *args, **kwargs)
    
class SerializerB(Serializer):
    encoding: str = None
    binary: bool = True
    read_mode: str = Mode.read_binary
    write_mode: str = Mode.write_binary
    append_mode: str = None
    default_mode: str = Mode.read_binary




import asyncio
import functools
import contextvars


async def _to_thread(func, *args, **kwargs):
    """Asynchronously run function *func* in a separate thread.
    Any *args and **kwargs supplied for this function are directly passed
    to *func*. Also, the current :class:`contextvars.Context` is propogated,
    allowing context variables from the main thread to be accessed in the
    separate thread.
    Return a coroutine that can be awaited to get the eventual result of *func*.
    """
    loop = asyncio.events.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)
    return await loop.run_in_executor(None, func_call)


def asyncify(obj, template: str = "async_[func_name]"):
    """
    Creates async function attributes for a class
    using the template
    i.e. obj.norm_call [sync] -> obj.async_norm_call [async]
    returns obj.
    
    AsyncCls = asyncify(SyncCls)

    a = AsyncCls.norm_call()
    x = await AsyncCls.async_norm_call()
    
    """
    def create_async_func(base_func):
        async def async_func(*args, **kwargs):
            return await _to_thread(base_func, *args, **kwargs)
        return async_func

    for attr in dir(obj):
        attr_val = getattr(obj, attr)
        if callable(attr_val):
            if not attr.startswith('_'):
                async_func = create_async_func(attr_val)
                setattr(obj, template.replace('[func_name]', attr), async_func)
        
    return obj


import time
from .types import *
from .models import BaseDataCls, BaseCls, Field


TimeValues = {
    'secs': 60,
    'mins': 60,
    'hrs': 60,
    'days': 24,
    'wks': 7,
    'mnths': 4
}


class TimeCls(BaseCls):
    t: Optional[Union[float, int]] = Field(default_factory=time.time)
    short: Optional[bool] = False
    start: Optional[Union[float, int]] = None

    @property
    def now(self):
        if self.start: return self.start
        return time.time()

    def stop(self): self.start = time.time()

    @property
    def diff(self): return self.now - self.t
    @property
    def s(self): return self.secs
    @property
    def secs(self): return self.diff
    @property
    def seconds(self): return self.secs
    @property
    def m(self): return self.mins
    @property
    def mins(self): return self.secs / 60
    @property
    def minutes(self): return self.mins
    @property
    def h(self): return self.hrs
    @property
    def hrs(self): return self.mins / 60
    @property
    def hr(self): return self.hrs
    @property
    def hour(self): return self.hrs
    @property
    def hours(self): return self.hrs
    @property
    def d(self): return self.days
    @property
    def day(self): return self.days
    @property
    def days(self): return self.hrs / 24
    @property
    def w(self): return self.wks
    @property
    def wk(self): return self.wks
    @property
    def wks(self): return self.days / 7
    @property
    def week(self): return self.wks
    @property
    def weeks(self): return self.wks
    @property
    def month(self): return self.mnths
    @property
    def mons(self): return self.mnths
    @property
    def mnths(self): return self.wks / 4
    @property
    def months(self): return self.mnths
    
    @property
    def ablstime(self) -> BaseDataCls:
        curr_val = self.secs
        dict_val, str_val = {}, ''
        for tkey, tnum in TimeValues.items():
            if curr_val >= 1:
                tval = tkey[0] if self.short else tkey
                curr_val, curr_num = divmod(curr_val, tnum)
                if type(curr_num) == float: str_val = f'{curr_num:.1f} {tval} ' + str_val
                else: str_val = f'{curr_num} {tval} ' + str_val
                dict_val[tkey] = curr_num
        str_val = str_val.strip()
        return BaseDataCls(string=str_val, value=dict_val, dtype='time')


__all__ = [
    'TimeCls'
]


import math

import sniffio
import threading
from importlib import import_module
from typing import Any, Callable, Coroutine, Dict, Generator, Optional, Tuple, Type, TypeVar

from lazy.libz import Lib

BACKENDS = 'asyncio', 'trio'

T_Retval = TypeVar('T_Retval')
threadlocals = threading.local()

DefaultAsyncBackend: str = 'trio' if Lib.is_avail_trio else 'asyncio'


"""
Borrowing from funcs and utils from anyio
"""

def run_async(func: Callable[..., Coroutine[Any, Any, T_Retval]], *args: object, backend: str = DefaultAsyncBackend, backend_options: Optional[Dict[str, Any]] = None) -> T_Retval:
    """
    Run the given coroutine function in an asynchronous event loop.
    The current thread must not be already running an event loop.
    :param func: a coroutine function
    :param args: positional arguments to ``func``
    :param backend: name of the asynchronous event loop implementation – currently either
        ``asyncio`` or ``trio``
    :param backend_options: keyword arguments to call the backend ``run()`` implementation with
        (documented :ref:`here <backend options>`)
    :return: the return value of the coroutine function
    :raises RuntimeError: if an asynchronous event loop is already running in this thread
    :raises LookupError: if the named backend is not found
    """
    try: asynclib_name = sniffio.current_async_library()
    except sniffio.AsyncLibraryNotFoundError: pass
    else: raise RuntimeError(f'Already running {asynclib_name} in this thread')

    try: asynclib = import_module(f'anyio._backends._{backend}', package=__name__)
    except ImportError as exc: raise LookupError(f'No such backend: {backend}') from exc

    token = None
    if sniffio.current_async_library_cvar.get(None) is None: token = sniffio.current_async_library_cvar.set(backend)

    try:
        backend_options = backend_options or {}
        return asynclib.run(func, *args, **backend_options)  # type: ignore
    finally:
        if token: sniffio.current_async_library_cvar.reset(token)






__all__ = (
    'Any',
    'AnyStr',
    'Dict',
    'Iterator',
    'List',
    'Optional',
    'Sequence',
    'Tuple',
    'Type',
    'TypeVar',
    'Union',
    'Protocol',
    'PathLike',
    'PathLikeCls',
    'T',
    'TupleOrList',
    'TreeDict',
    'Tree',
    'Tensor',
    'Dim',
    'Shape',
    'JsonValue',
    'JsonType',
    'Key',
    'KeySerializedExample',
    'ModuleType',
)

import os
from typing import Any, AnyStr, Dict, Iterator, List, Optional, Sequence, Tuple, Type, TypeVar, Union
from types import ModuleType

try:
    from typing import Protocol
except ImportError:
    import typing_extensions
    Protocol = typing_extensions.Protocol

# Accept both `str` and `pathlib.Path`-like
PathLike = Union[str, os.PathLike]
PathLikeCls = (str, os.PathLike)    # Used in `isinstance`

T = TypeVar('T')

# Note: `TupleOrList` avoid abiguity from `Sequence` (`str` is `Sequence[str]`,
# `bytes` is `Sequence[int]`).
TupleOrList = Union[Tuple[T, ...], List[T]]

TreeDict = Union[T, Dict[str, 'TreeDict']]    # pytype: disable=not-supported-yet
Tree = Union[T, TupleOrList['Tree'], Dict[str, 'Tree']]    # pytype: disable=not-supported-yet

Tensor = Union[T, Any]

Dim = Optional[int]
Shape = TupleOrList[Dim]

JsonValue = Union[str, bool, int, float, None, List['JsonValue'], Dict[str, 'JsonValue']]
JsonType = Dict[str, JsonValue]

Key = Union[int, str, bytes]
KeySerializedExample = Tuple[Key, bytes]






"""
Cloud Fuze Implementations

libfuse-dev
libfuse3-dev
fuse3

# -Displaymode 0
"""


from typing import Optional
from .core import BaseFuzerCls


class GCSFuze(BaseFuzerCls):
    """
    GCS Fuze
    """
    _FSX_LIB: str = 'gcsfs'
    _FSX_MODULE: Optional[str] = None
    _FSX_CLS: str = 'GCSFileSystem'

    @classmethod
    def get_configz(cls, reload: bool = False, **config):
        authz = cls.get_authz(reload=reload, **config)
        _config = {}
        if authz.gcp_auth: _config['token'] = authz.gcp_auth
        if authz.gcloud_project or authz.google_cloud_project: _config['project'] = authz.gcloud_project or authz.google_cloud_project
        if authz.gcs_client_config: _config['client_kwargs'] = authz.gcs_client_config
        if authz.gcs_config: _config['config_kwargs'] = authz.gcs_config
        return _config

class S3Fuze(BaseFuzerCls):
    """
    S3 Fuze
    """
    _FSX_LIB: str = 's3fs'
    _FSX_MODULE: Optional[str] = None
    _FSX_CLS: str = 'S3FileSystem'

    @classmethod
    def get_configz(cls, reload: bool = False, **config):
        authz = cls.get_authz(reload=reload, **config)
        _config = {}
        if authz.aws_access_key_id:
            _config['key'] = authz.aws_access_key_id
            _config['secret'] = authz.aws_secret_access_key
        
        elif authz.aws_access_token: _config['token'] = authz.aws_access_token
        elif not authz.boto_config: _config['anon'] = True
        if authz.aws_region: _config['client_kwargs'] = {'endpoint_url': authz.get_s3_endpoint(), 'region_name': authz.aws_region}
        if authz.s3_config: _config['config_kwargs'] = authz.s3_config
        return _config


class MinioFuze(BaseFuzerCls):
    """
    Minio Fuze based on S3
    """
    _FSX_LIB: str = 's3fs'
    _FSX_MODULE: Optional[str] = None
    _FSX_CLS: str = 'S3FileSystem'

    @classmethod
    def get_configz(cls, reload: bool = False, **config):
        authz = cls.get_authz(reload=reload, **config)
        _config = {}
        if authz.minio_secret_key:
            _config['key'] = authz.minio_access_key
            _config['secret'] = authz.minio_secret_key
        elif authz.minio_access_token: _config['token'] = authz.minio_access_token
        _config['client_kwargs'] = {'endpoint_url': authz.minio_endpoint}
        if authz.minio_config: _config['config_kwargs'] = authz.minio_config
        return _config

    

class S3CompatFuze(BaseFuzerCls):
    """
    S3Compatible Generic Fuze based for providers like digitalocean, etc.
    """
    _FSX_LIB: str = 's3fs'
    _FSX_MODULE: Optional[str] = None
    _FSX_CLS: str = 'S3FileSystem'

    @classmethod
    def get_configz(cls, reload: bool = False, **config):
        authz = cls.get_authz(reload=reload, **config)
        _config = {}
        if authz.s3compat_secret_key:
            _config['key'] = authz.s3compat_access_key
            _config['secret'] = authz.s3compat_secret_key
        elif authz.s3compat_access_token: _config['token'] = authz.s3compat_access_token
        _config['client_kwargs'] = {'endpoint_url': authz.s3compat_endpoint}
        if authz.s3compat_region: _config['client_kwargs']['region_name'] = authz.s3compat_region
        if authz.s3compat_config: _config['config_kwargs'] = authz.s3compat_config
        return _config



import gzip as _gzip
import zlib as _zlib
import bz2 as _bz2
import pickle as _pickle
import dill as _dill

from typing import Any
from .core import SerializerB


class DefaultProtocols:
    default: int = 4
    pickle: int = _pickle.HIGHEST_PROTOCOL
    dill: int = _dill.HIGHEST_PROTOCOL

class DefaultCompression:
    default: int = -1
    gzip: int = 10
    zlib: int = 6
    bz2: int = 9


class Compression:
    gzip = _gzip
    zlib = _zlib
    bz2 = _bz2


class BasePickle(SerializerB):
    default_value: Any = None
    async_supported: bool = True
    cloud_supported: bool = True
    compression_supported: bool = True
    
    @classmethod
    def dumps(cls, obj: Any, protocol: int = DefaultProtocols.default, compress: bool = False, *args, default: Any = None, **kwargs) -> bytes:
        return cls._encode(obj, protocol = protocol, compress = compress, *args, default = default, **kwargs)

    @classmethod
    def loads(cls, data: bytes, decompress: bool = False, *args, **kwargs) -> Any:
        return cls._decode(data, decompress = decompress, *args, **kwargs)
    
    @classmethod
    async def async_dumps(cls, obj: Any, protocol: int = DefaultProtocols.default, compress: bool = False, *args, default: Any = None, **kwargs) -> bytes:
        if not cls.async_supported: raise Exception
        return await cls._async_encode(obj, protocol = protocol, compress = compress , *args, default = default, **kwargs)

    @classmethod
    async def async_loads(cls, data: bytes, decompress: bool = False, *args, **kwargs) -> Any:
        if not cls.async_supported: raise Exception
        return await cls._async_decode(data, decompress = decompress, *args, **kwargs)
    
    @classmethod
    def compress(cls, data: bytes, method: str = 'gzip', compressionlvl: int = DefaultCompression.default, *args, **kwargs) -> bytes:
        _method = getattr(Compression, method)
        return _method.compress(data, compressionlvl = compressionlvl, *args, **kwargs)
    
    @classmethod
    def decompress(cls, data: bytes, method: str = 'gzip', *args, **kwargs) -> bytes:
        _method = getattr(Compression, method)
        return _method.decompress(data, *args, **kwargs)

    @classmethod
    def _encode(cls, obj: Any, protocol: int = DefaultProtocols.default, compress: bool = False, *args, default: Any = None, **kwargs) -> bytes:
        raise NotImplementedError
        
    @classmethod
    def _decode(cls, data: bytes, decompress: bool = False, *args, **kwargs) -> Any:
        raise NotImplementedError
    
    @classmethod
    async def _async_encode(cls, obj: Any, protocol: int = DefaultProtocols.default, compress: bool = False, *args, default: Any = None, **kwargs) -> bytes:
        raise NotImplementedError
    
    @classmethod
    async def _async_decode(cls, data: bytes, decompress: bool = False, *args, **kwargs) -> Any:
        raise NotImplementedError


class Pickle(BasePickle):
    @classmethod
    def _encode(cls, obj: Any, protocol: int = DefaultProtocols.pickle, compress: bool = False, *args, default: Any = None, **kwargs) -> bytes:
        data = _pickle.dumps(obj, protocol=protocol, *args, **kwargs)
        if compress: data = cls.compress(data, *args, **kwargs)
        return data
        
    @classmethod
    def _decode(cls, data: bytes, decompress: bool = False, *args, **kwargs) -> Any:
        if decompress: data = cls.decompress(data, *args, **kwargs)
        return _pickle.loads(data, *args, **kwargs)
          
    @classmethod
    async def _async_encode(cls, obj: Any, protocol: int = DefaultProtocols.pickle, compress: bool = False, *args, default: Any = None, **kwargs) -> bytes:
        data = _pickle.dumps(obj, protocol=protocol, *args, **kwargs)
        if compress: data = cls.compress(data, *args, **kwargs)
        return data
    
    @classmethod
    async def _async_decode(cls, data: bytes, decompress: bool = False, *args, **kwargs) -> Any:
        if decompress: data = cls.decompress(data, *args, **kwargs)
        return _pickle.loads(data, *args, **kwargs)


class Dill(BasePickle):
    @classmethod
    def _encode(cls, obj: Any, protocol: int = DefaultProtocols.dill, compress: bool = False, *args, default: Any = None, **kwargs) -> bytes:
        data = _dill.dumps(obj, protocol=protocol, *args, **kwargs)
        if compress: data = cls.compress(data, *args, **kwargs)
        return data
        
    @classmethod
    def _decode(cls, data: bytes, decompress: bool = False, *args, **kwargs) -> Any:
        if decompress: data = cls.decompress(data, *args, **kwargs)
        return _dill.loads(data, *args, **kwargs)
          
    @classmethod
    async def _async_encode(cls, obj: Any, protocol: int = DefaultProtocols.dill, compress: bool = False, *args, default: Any = None, **kwargs) -> bytes:
        data = _dill.dumps(obj, protocol=protocol, *args, **kwargs)
        if compress: data = cls.compress(data, *args, **kwargs)
        return data
    
    @classmethod
    async def _async_decode(cls, data: bytes, decompress: bool = False, *args, **kwargs) -> Any:
        if decompress: data = cls.decompress(data, *args, **kwargs)
        return _dill.loads(data, *args, **kwargs)

class Pkl(Dill):
    pass



import base64
import gzip
import hashlib
import uuid

from typing import Dict, Any, List, Union, overload
from .base import Srlzer, Mode

DEFAULT_BASE_METHOD = 'base64'

class Base(Srlzer):
    default_value: str = None
    async_supported: bool = True
    cloud_supported: bool = True
    hash_method: str = "sha256"

    @classmethod
    def b64_encode(cls, text: str) -> str:
        return base64.b64encode(text.encode(encoding=cls.encoding)).decode(encoding=cls.encoding)

    @classmethod
    def b64_decode(cls, data: Union[str, bytes]) -> str:
        if isinstance(data, str): data = data.encode(encoding=cls.encoding)
        return base64.b64decode(data).decode(encoding=cls.encoding)

    @classmethod
    def b64_gzip_encode(cls, text: str) -> str:
        return base64.b64encode(gzip.compress(text.encode(encoding=cls.encoding))).decode(encoding=cls.encoding)

    @classmethod
    def b64_gzip_decode(cls, data: Union[str, bytes]) -> str:
        if isinstance(data, str): data = data.encode(encoding=cls.encoding)
        return gzip.decompress(base64.b64decode(data)).decode(encoding=cls.encoding)

    @classmethod
    def hash_encode(cls, text: str, method: str = 'sha256') -> str:
        encoder = getattr(hashlib, method)
        return encoder(text.encode(encoding=cls.encoding)).hexdigest()

    @classmethod
    def hash_compare(cls, text: str, hashtext: str, method: str = 'sha256') -> bool:
        return bool(cls.hash_encode(text, method) == hashtext)

    @classmethod
    def hash_b64_encode(cls, text: str, method: str = 'sha256') -> str:
        return cls.hash_encode(cls.b64_encode(text), method=method)

    @classmethod
    def hash_b64_gzip_encode(cls, text: str, method: str = 'sha256') -> str:
        return cls.hash_encode(cls.b64_gzip_encode(text), method=method)

    @classmethod
    def hash_b64_compare(cls, hashtext: str, data: Union[str, bytes], method: str = 'sha256') -> bool:
        if isinstance(data, str): return bool(cls.hash_b64_encode(text=data, method=method) == hashtext)
        return bool(cls.b64_decode(data) == hashtext)

    @classmethod
    def hash_b64_gzip_compare(cls, hashtext: str, data: Union[str, bytes], method: str = 'sha256') -> bool:
        if isinstance(data, str): return bool(cls.hash_b64_gzip_encode(text=data, method=method) == hashtext)
        return bool(cls.b64_gzip_decode(data) == hashtext)

    @classmethod
    def hash_b64_compare_match(cls, text: str, data: Union[str, bytes], method: str = 'sha256') -> bool:
        if isinstance(data, str): data = data.encode(encoding=cls.encoding)
        return bool(cls.b64_decode(data) == cls.hash_encode(text, method=method))

    @classmethod
    def hash_b64_gzip_compare_match(cls, text: str, data: Union[str, bytes], method: str = 'sha256') -> bool:
        if isinstance(data, str): data = data.encode(encoding=cls.encoding)
        return bool(cls.b64_gzip_decode(data) == cls.hash_encode(text, method=method))

    @staticmethod
    def get_uuid(method: str = 'uuid4', *args, **kwargs):
        t = getattr(uuid, method, 'uuid4')
        return str(t(*args, **kwargs))
    
    @classmethod
    def get_encode_method(cls, method: str = DEFAULT_BASE_METHOD):
        if method.lower() in {'b64', 'base64'}: return cls.b64_encode
        if method.lower() in {'bgz', 'base64gzip'}: return cls.b64_gzip_encode
        if method.lower() in {'hash', 'h'}: return cls.hash_encode
        raise NotImplementedError
    
    @classmethod
    def get_decode_method(cls, method: str = DEFAULT_BASE_METHOD):
        if method.lower() in {'b64', 'base64'}: return cls.b64_decode
        if method.lower() in {'bgz', 'base64gzip'}: return cls.b64_gzip_decode
        raise NotImplementedError
    
    @classmethod
    async def async_dumps(cls, data: str, method: str = DEFAULT_BASE_METHOD, *args, **kwargs):
        return await cls._async_encode(data, method = method, *args, **kwargs)

    @classmethod
    async def async_loads(cls, data: str, method: str = DEFAULT_BASE_METHOD, *args, **kwargs):
        return await cls._async_decode(data, method = method, *args, **kwargs)
    
    @classmethod
    def dumps(cls, data: str, method: str = DEFAULT_BASE_METHOD, *args, **kwargs) -> str:
        return cls._encode(data, method = method, *args, **kwargs)

    @classmethod
    def loads(cls, data: Union[str, bytes], method: str = DEFAULT_BASE_METHOD, *args, **kwargs) -> str:
        return cls._decode(data, method = method, *args, **kwargs)
    
    @classmethod
    def _encode(cls, data: str, method: str = DEFAULT_BASE_METHOD, *args,  **kwargs) -> str:
        _method = cls.get_encode_method(method)
        return _method(data, *args, **kwargs)

    @classmethod
    def _decode(cls, data: Union[str, bytes], method: str = DEFAULT_BASE_METHOD, *args, **kwargs) -> str:
        _method = cls.get_decode_method(method)
        return _method(data, *args, **kwargs)
    
    @classmethod
    async def _async_encode(cls, data: str, method: str = DEFAULT_BASE_METHOD, *args, **kwargs) -> str:
        _method = cls.get_encode_method(method)
        return _method(data, *args, **kwargs)
    
    @classmethod
    async def _async_decode(cls, data: Union[str, bytes], method: str = DEFAULT_BASE_METHOD, *args, **kwargs) -> str:
        _method = cls.get_decode_method(method)
        return _method(data, *args, **kwargs)

        
    



"""
Pydantic Types
"""

__all__ = (
    'NoneStr',
    'NoneBytes',
    'StrBytes',
    'NoneStrBytes',
    'StrictStr',
    'ConstrainedBytes',
    'conbytes',
    'ConstrainedList',
    'conlist',
    'ConstrainedSet',
    'conset',
    'ConstrainedStr',
    'constr',
    'PyObject',
    'ConstrainedInt',
    'conint',
    'PositiveInt',
    'NegativeInt',
    'NonNegativeInt',
    'NonPositiveInt',
    'ConstrainedFloat',
    'confloat',
    'PositiveFloat',
    'NegativeFloat',
    'NonNegativeFloat',
    'NonPositiveFloat',
    'ConstrainedDecimal',
    'condecimal',
    'UUID1',
    'UUID3',
    'UUID4',
    'UUID5',
    'FilePath',
    'DirectoryPath',
    'Json',
    'JsonWrapper',
    'SecretStr',
    'SecretBytes',
    'StrictBool',
    'StrictBytes',
    'StrictInt',
    'StrictFloat',
    'PaymentCardNumber',
    'ByteSize'
)

from pydantic.types import *




import yaml

from typing import Dict, Any, List, Union, overload
from .base import Srlzer, Mode


class Yaml(Srlzer):
    default_value: Dict[Any, Any] = None
    async_supported: bool = True
    cloud_supported: bool = True
    
    
    @classmethod
    def dumps(cls, obj: Dict[Any, Any], dumper: str = 'default', *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        return cls._encode(obj, dumper = dumper, *args, default = default, **kwargs)

    @classmethod
    def loads(cls, obj: Union[str, bytes], loader: str = 'default', *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        return cls._decode(obj, loader = loader, *args, **kwargs)
    
    @classmethod
    def _encode(cls, obj: Dict[Any, Any], dumper: str = 'default', *args, default: Dict[Any, Any] = None, **kwargs) -> str: 
        _dumper = yaml.Dumper
        if dumper == 'safe': _dumper = yaml.SafeDumper
        return yaml.dump(obj, Dumper=_dumper *args, **kwargs)

    @classmethod
    def _decode(cls, obj: Union[str, bytes], loader: str = 'default', *args, **kwargs) -> Union[Dict[Any, Any], List[str]]: 
        _loader = yaml.Loader
        if loader == 'safe': _loader = yaml.SafeLoader
        return yaml.load(obj, Loader=_loader, *args, **kwargs)
    
    @classmethod
    async def async_dumps(cls, obj: Dict[Any, Any], dumper: str = 'default', *args, default: Any = None, **kwargs) -> str:
        if not cls.async_supported: raise Exception
        return await cls._async_encode(obj, dumper = dumper, *args, default = default, **kwargs)

    @classmethod
    async def async_loads(cls, obj: Union[str, bytes], loader: str = 'default', *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if not cls.async_supported: raise Exception
        return await cls._async_decode(obj, loader = loader, *args, **kwargs)
    
    @overload
    @classmethod
    async def _async_encode(cls, obj: Dict[Any, Any], dumper: str = 'default', *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        _dumper = yaml.Dumper
        if dumper == 'safe': _dumper = yaml.SafeDumper
        return yaml.dump(obj, Dumper=_dumper *args, **kwargs)
    
    @overload
    @classmethod
    async def _async_decode(cls, obj: Union[str, bytes], loader: str = 'default', *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        _loader = yaml.Loader
        if loader == 'safe': _loader = yaml.SafeLoader
        return yaml.load(obj, Loader=_loader, *args, **kwargs)


import time

from logz import get_logger, get_cls_logger
from lazy.static.timez import TimeValues
from lazy.models import BaseCls

class TimeResult(BaseCls):
    secs: float = 0.0
    mins: float = 0.0
    hrs: float = 0.0
    days: float = 0.0
    weeks: float = 0.0
    months: float = 0.0

    def get_string(self, short: bool = False, ksize: int = 4, nfloat: int = 5):
        s = ''
        d = self.dict()
        for key in {'secs', 'mins', 'hrs', 'days', 'weeks', 'months'}:
            if d.get(key):
                v = str(round(d[key], nfloat))
                k = key[:ksize] if short else key
                s += f' {v} {k}'
        return s.strip()

    @property
    def string(self):
        return self.get_string()

    @property
    def short(self):
        return self.get_string(short=True)


def to_time_string(t: float, short: bool = False, as_obj: bool = False):
    dict_val = {}
    curr_val = t
    for tkey, tnum in TimeValues.items():
        if curr_val > 0:
            curr_val, curr_num = divmod(curr_val, tnum)
            dict_val[tkey] = curr_num
    tcls = TimeResult(**dict_val)
    if as_obj: return tcls
    if short: return tcls.short
    return tcls.string



def timer(start: float = None, as_string: bool = False, short: bool = False, as_obj: bool = False, **kwargs):
    """
    Simple Timer Function that returns time.time 
    or result of start - now
    """
    if not start: return time.time()
    stop = (time.time() - start)
    if not as_string: return stop
    return to_time_string(stop, short=short, as_obj = as_obj)





""" 
Source: https://github.com/grantjenks/python-diskcache/blob/71db7753610bcd4bd7adda28abeb2b0fdebbc31f/diskcache/core.py
"""

import codecs
import contextlib as cl
import errno
import functools as ft
import inspect
import io
import json
import os
import os.path as op
import pickletools
import sqlite3
import struct
import tempfile
import threading
import time
import warnings
import zlib
import dill

from .static import *

_zlib = zlib
_isal_attempted: bool = False

def get_zlib():
    global _zlib, _isal_attempted
    if _isal_attempted: return
    from lazy.libz import Lib
    try:
        _isal = Lib.import_lib('isal')
        _zlib = _isal.isal_zlib #import_module('isal_zlib', 'isal')
    except Exception as e: 
        #print(e)
        pass
    _isal_attempted = True




class Disk:
    "Cache key and value serialization for SQLite database and files."

    def __init__(self, directory, min_file_size=0, pickle_protocol=0):
        """Initialize disk instance.
        :param str directory: directory path
        :param int min_file_size: minimum size for file use
        :param int pickle_protocol: pickle protocol for serialization
        """
        get_zlib()
        self._directory = directory
        self.min_file_size = min_file_size
        self.pickle_protocol = pickle_protocol

    def hash(self, key):
        """Compute portable hash for `key`.
        :param key: key to hash
        :return: hash value
        """
        mask = 0xFFFFFFFF
        disk_key, _ = self.put(key)
        type_disk_key = type(disk_key)

        if type_disk_key is sqlite3.Binary:
            return _zlib.adler32(disk_key) & mask
        elif type_disk_key is str:
            return _zlib.adler32(disk_key.encode('utf-8')) & mask  # noqa
        elif type_disk_key is int:
            return disk_key % mask
        else:
            assert type_disk_key is float
            return _zlib.adler32(struct.pack('!d', disk_key)) & mask

    def put(self, key):
        """Convert `key` to fields key and raw for Cache table.
        :param key: key to convert
        :return: (database key, raw boolean) pair
        """
        # pylint: disable=unidiomatic-typecheck
        type_key = type(key)

        if type_key is bytes:
            return sqlite3.Binary(key), True
        elif (
            (type_key is str)
            or (
                type_key is int
                and -9223372036854775808 <= key <= 9223372036854775807
            )
            or (type_key is float)
        ):
            return key, True
        else:
            data = dill.dumps(key, protocol=self.pickle_protocol)
            result = pickletools.optimize(data)
            return sqlite3.Binary(result), False

    def get(self, key, raw):
        """Convert fields `key` and `raw` from Cache table to key.
        :param key: database key to convert
        :param bool raw: flag indicating raw database storage
        :return: corresponding Python key
        """
        # pylint: disable=no-self-use,unidiomatic-typecheck
        if raw:
            return bytes(key) if type(key) is sqlite3.Binary else key
        else:
            return dill.load(io.BytesIO(key))

    def store(self, value, read, key=UNKNOWN):
        """Convert `value` to fields size, mode, filename, and value for Cache
        table.
        :param value: value to convert
        :param bool read: True when value is file-like object
        :param key: key for item (default UNKNOWN)
        :return: (size, mode, filename, value) tuple for Cache table
        """
        # pylint: disable=unidiomatic-typecheck
        type_value = type(value)
        min_file_size = self.min_file_size

        if (
            (type_value is str and len(value) < min_file_size)
            or (
                type_value is int
                and -9223372036854775808 <= value <= 9223372036854775807
            )
            or (type_value is float)
        ):
            return 0, MODE_RAW, None, value
        elif type_value is bytes:
            if len(value) < min_file_size:
                return 0, MODE_RAW, None, sqlite3.Binary(value)
            else:
                filename, full_path = self.filename(key, value)

                with open(full_path, 'xb') as writer:
                    writer.write(value)

                return len(value), MODE_BINARY, filename, None
        elif type_value is str:
            filename, full_path = self.filename(key, value)

            with open(full_path, 'x', encoding='UTF-8') as writer:
                writer.write(value)

            size = op.getsize(full_path)
            return size, MODE_TEXT, filename, None
        elif read:
            size = 0
            reader = ft.partial(value.read, 2 ** 22)
            filename, full_path = self.filename(key, value)

            with open(full_path, 'xb') as writer:
                for chunk in iter(reader, b''):
                    size += len(chunk)
                    writer.write(chunk)

            return size, MODE_BINARY, filename, None
        else:
            result = dill.dumps(value, protocol=self.pickle_protocol)

            if len(result) < min_file_size:
                return 0, MODE_PICKLE, None, sqlite3.Binary(result)
            else:
                filename, full_path = self.filename(key, value)

                with open(full_path, 'xb') as writer:
                    writer.write(result)

                return len(result), MODE_PICKLE, filename, None

    def fetch(self, mode, filename, value, read):
        """Convert fields `mode`, `filename`, and `value` from Cache table to
        value.
        :param int mode: value mode raw, binary, text, or pickle
        :param str filename: filename of corresponding value
        :param value: database value
        :param bool read: when True, return an open file handle
        :return: corresponding Python value
        """
        # pylint: disable=no-self-use,unidiomatic-typecheck
        if mode == MODE_RAW:
            return bytes(value) if type(value) is sqlite3.Binary else value
        elif mode == MODE_BINARY:
            if read:
                return open(op.join(self._directory, filename), 'rb')
            else:
                with open(op.join(self._directory, filename), 'rb') as reader:
                    return reader.read()
        elif mode == MODE_TEXT:
            full_path = op.join(self._directory, filename)
            with open(full_path, 'r', encoding='UTF-8') as reader:
                return reader.read()
        elif mode == MODE_PICKLE:
            if value is None:
                with open(op.join(self._directory, filename), 'rb') as reader:
                    return dill.load(reader)
            else:
                return dill.load(io.BytesIO(value))

    def filename(self, key=UNKNOWN, value=UNKNOWN):
        """Return filename and full-path tuple for file storage.
        Filename will be a randomly generated 28 character hexadecimal string
        with ".val" suffixed. Two levels of sub-directories will be used to
        reduce the size of directories. On older filesystems, lookups in
        directories with many files may be slow.
        The default implementation ignores the `key` and `value` parameters.
        In some scenarios, for example :meth:`Cache.push
        <diskcache.Cache.push>`, the `key` or `value` may not be known when the
        item is stored in the cache.
        :param key: key for item (default UNKNOWN)
        :param value: value for item (default UNKNOWN)
        """
        # pylint: disable=unused-argument
        hex_name = codecs.encode(os.urandom(16), 'hex').decode('utf-8')
        sub_dir = op.join(hex_name[:2], hex_name[2:4])
        name = hex_name[4:] + '.val'
        directory = op.join(self._directory, sub_dir)

        try:
            os.makedirs(directory)
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise

        filename = op.join(sub_dir, name)
        full_path = op.join(self._directory, filename)
        return filename, full_path

    def remove(self, filename):
        """Remove a file given by `filename`.
        This method is cross-thread and cross-process safe. If an "error no
        entry" occurs, it is suppressed.
        :param str filename: relative path to file
        """
        full_path = op.join(self._directory, filename)

        try:
            os.remove(full_path)
        except WindowsError:
            pass
        except OSError as error:
            if error.errno != errno.ENOENT:
                # ENOENT may occur if two caches attempt to delete the same
                # file at the same time.
                raise



class JSONDisk(Disk):
    "Cache key and value using JSON serialization with zlib compression."

    def __init__(self, directory, compress_level=3, **kwargs):
        """Initialize JSON disk instance.
        Keys and values are compressed using the zlib library. The
        `compress_level` is an integer from 0 to 9 controlling the level of
        compression; 1 is fastest and produces the least compression, 9 is
        slowest and produces the most compression, and 0 is no compression.
        :param str directory: directory path
        :param int compress_level: zlib compression level (default 1)
        :param kwargs: super class arguments
        """
        self.compress_level = compress_level
        super().__init__(directory, **kwargs)

    def put(self, key):
        json_bytes = json.dumps(key).encode('utf-8')
        data = zlib.compress(json_bytes, self.compress_level)
        return super().put(data)

    def get(self, key, raw):
        data = super().get(key, raw)
        return json.loads(zlib.decompress(data).decode('utf-8'))

    def store(self, value, read, key=UNKNOWN):
        if not read:
            json_bytes = json.dumps(value).encode('utf-8')
            value = zlib.compress(json_bytes, self.compress_level)
        return super().store(value, read, key=key)

    def fetch(self, mode, filename, value, read):
        data = super().fetch(mode, filename, value, read)
        if not read:
            data = json.loads(zlib.decompress(data).decode('utf-8'))
        return data


class Timeout(Exception):
    "Database timeout expired."


class UnknownFileWarning(UserWarning):
    "Warning used by Cache.check for unknown files."


class EmptyDirWarning(UserWarning):
    "Warning used by Cache.check for empty directories."


def args_to_key(base, args, kwargs, typed):
    """Create cache key out of function arguments.
    :param tuple base: base of key
    :param tuple args: function arguments
    :param dict kwargs: function keyword arguments
    :param bool typed: include types in cache key
    :return: cache key tuple
    """
    key = base + args

    if kwargs:
        key += (ENOVAL,)
        sorted_items = sorted(kwargs.items())

        for item in sorted_items:
            key += item

    if typed:
        key += tuple(type(arg) for arg in args)

        if kwargs:
            key += tuple(type(value) for _, value in sorted_items)

    return key


class Cache:
    "Disk and file backed cache."

    def __init__(self, directory=None, timeout=60, disk=Disk, **settings):
        """Initialize cache instance.
        :param str directory: cache directory
        :param float timeout: SQLite connection timeout
        :param disk: Disk type or subclass for serialization
        :param settings: any of DEFAULT_SETTINGS
        """
        try:
            assert issubclass(disk, Disk)
        except (TypeError, AssertionError):
            raise ValueError('disk must subclass lazy.io.cachez.Disk') from None
        if directory is None:
            directory = tempfile.mkdtemp(prefix='cachez-')
        directory = op.expanduser(directory)
        directory = op.expandvars(directory)

        self._directory = directory
        self._timeout = 0  # Manually handle retries during initialization.
        self._local = threading.local()
        self._txn_id = None

        if not op.isdir(directory):
            try:
                os.makedirs(directory, 0o755)
            except OSError as error:
                if error.errno != errno.EEXIST:
                    raise EnvironmentError(
                        error.errno,
                        'Cache directory "%s" does not exist'
                        ' and could not be created' % self._directory,
                    ) from None

        sql = self._sql_retry

        # Setup Settings table.

        try:
            current_settings = dict(
                sql('SELECT key, value FROM Settings').fetchall()
            )
        except sqlite3.OperationalError:
            current_settings = {}

        sets = DEFAULT_SETTINGS.copy()
        sets.update(current_settings)
        sets.update(settings)

        for key in METADATA:
            sets.pop(key, None)

        # Chance to set pragmas before any tables are created.

        for key, value in sorted(sets.items()):
            if key.startswith('sqlite_'):
                self.reset(key, value, update=False)

        sql(
            'CREATE TABLE IF NOT EXISTS Settings ('
            ' key TEXT NOT NULL UNIQUE,'
            ' value)'
        )

        # Setup Disk object (must happen after settings initialized).

        kwargs = {
            key[5:]: value
            for key, value in sets.items()
            if key.startswith('disk_')
        }
        self._disk = disk(directory, **kwargs)

        # Set cached attributes: updates settings and sets pragmas.

        for key, value in sets.items():
            query = 'INSERT OR REPLACE INTO Settings VALUES (?, ?)'
            sql(query, (key, value))
            self.reset(key, value)

        for key, value in METADATA.items():
            query = 'INSERT OR IGNORE INTO Settings VALUES (?, ?)'
            sql(query, (key, value))
            self.reset(key)

        ((self._page_size,),) = sql('PRAGMA page_size').fetchall()

        # Setup Cache table.

        sql(
            'CREATE TABLE IF NOT EXISTS Cache ('
            ' rowid INTEGER PRIMARY KEY,'
            ' key BLOB,'
            ' raw INTEGER,'
            ' store_time REAL,'
            ' expire_time REAL,'
            ' access_time REAL,'
            ' access_count INTEGER DEFAULT 0,'
            ' tag BLOB,'
            ' size INTEGER DEFAULT 0,'
            ' mode INTEGER DEFAULT 0,'
            ' filename TEXT,'
            ' value BLOB)'
        )

        sql(
            'CREATE UNIQUE INDEX IF NOT EXISTS Cache_key_raw ON'
            ' Cache(key, raw)'
        )

        sql(
            'CREATE INDEX IF NOT EXISTS Cache_expire_time ON'
            ' Cache (expire_time)'
        )

        query = EVICTION_POLICY[self.eviction_policy]['init']

        if query is not None:
            sql(query)

        # Use triggers to keep Metadata updated.

        sql(
            'CREATE TRIGGER IF NOT EXISTS Settings_count_insert'
            ' AFTER INSERT ON Cache FOR EACH ROW BEGIN'
            ' UPDATE Settings SET value = value + 1'
            ' WHERE key = "count"; END'
        )

        sql(
            'CREATE TRIGGER IF NOT EXISTS Settings_count_delete'
            ' AFTER DELETE ON Cache FOR EACH ROW BEGIN'
            ' UPDATE Settings SET value = value - 1'
            ' WHERE key = "count"; END'
        )

        sql(
            'CREATE TRIGGER IF NOT EXISTS Settings_size_insert'
            ' AFTER INSERT ON Cache FOR EACH ROW BEGIN'
            ' UPDATE Settings SET value = value + NEW.size'
            ' WHERE key = "size"; END'
        )

        sql(
            'CREATE TRIGGER IF NOT EXISTS Settings_size_update'
            ' AFTER UPDATE ON Cache FOR EACH ROW BEGIN'
            ' UPDATE Settings'
            ' SET value = value + NEW.size - OLD.size'
            ' WHERE key = "size"; END'
        )

        sql(
            'CREATE TRIGGER IF NOT EXISTS Settings_size_delete'
            ' AFTER DELETE ON Cache FOR EACH ROW BEGIN'
            ' UPDATE Settings SET value = value - OLD.size'
            ' WHERE key = "size"; END'
        )

        # Create tag index if requested.

        if self.tag_index:  # pylint: disable=no-member
            self.create_tag_index()
        else:
            self.drop_tag_index()

        # Close and re-open database connection with given timeout.

        self.close()
        self._timeout = timeout
        self._sql  # pylint: disable=pointless-statement

    @property
    def directory(self):
        """Cache directory."""
        return self._directory

    @property
    def timeout(self):
        """SQLite connection timeout value in seconds."""
        return self._timeout

    @property
    def disk(self):
        """Disk used for serialization."""
        return self._disk

    @property
    def _con(self):
        # Check process ID to support process forking. If the process
        # ID changes, close the connection and update the process ID.

        local_pid = getattr(self._local, 'pid', None)
        pid = os.getpid()

        if local_pid != pid:
            self.close()
            self._local.pid = pid

        con = getattr(self._local, 'con', None)

        if con is None:
            con = self._local.con = sqlite3.connect(
                op.join(self._directory, DBNAME),
                timeout=self._timeout,
                isolation_level=None,
            )

            # Some SQLite pragmas work on a per-connection basis so
            # query the Settings table and reset the pragmas. The
            # Settings table may not exist so catch and ignore the
            # OperationalError that may occur.

            try:
                select = 'SELECT key, value FROM Settings'
                settings = con.execute(select).fetchall()
            except sqlite3.OperationalError:
                pass
            else:
                for key, value in settings:
                    if key.startswith('sqlite_'):
                        self.reset(key, value, update=False)

        return con

    @property
    def _sql(self):
        return self._con.execute

    @property
    def _sql_retry(self):
        sql = self._sql

        # 2018-11-01 GrantJ - Some SQLite builds/versions handle
        # the SQLITE_BUSY return value and connection parameter
        # "timeout" differently. For a more reliable duration,
        # manually retry the statement for 60 seconds. Only used
        # by statements which modify the database and do not use
        # a transaction (like those in ``__init__`` or ``reset``).
        # See Issue #85 for and tests/issue_85.py for more details.

        def _execute_with_retry(statement, *args, **kwargs):
            start = time.time()
            while True:
                try:
                    return sql(statement, *args, **kwargs)
                except sqlite3.OperationalError as exc:
                    if str(exc) != 'database is locked':
                        raise
                    diff = time.time() - start
                    if diff > 60:
                        raise
                    time.sleep(0.001)

        return _execute_with_retry

    @cl.contextmanager
    def transact(self, retry=False):
        """Context manager to perform a transaction by locking the cache.
        While the cache is locked, no other write operation is permitted.
        Transactions should therefore be as short as possible. Read and write
        operations performed in a transaction are atomic. Read operations may
        occur concurrent to a transaction.
        Transactions may be nested and may not be shared between threads.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        >>> cache = Cache()
        >>> with cache.transact():  # Atomically increment two keys.
        ...     _ = cache.incr('total', 123.4)
        ...     _ = cache.incr('count', 1)
        >>> with cache.transact():  # Atomically calculate average.
        ...     average = cache['total'] / cache['count']
        >>> average
        123.4
        :param bool retry: retry if database timeout occurs (default False)
        :return: context manager for use in `with` statement
        :raises Timeout: if database timeout occurs
        """
        with self._transact(retry=retry):
            yield

    @cl.contextmanager
    def _transact(self, retry=False, filename=None):
        sql = self._sql
        filenames = []
        _disk_remove = self._disk.remove
        tid = threading.get_ident()
        txn_id = self._txn_id

        if tid == txn_id:
            begin = False
        else:
            while True:
                try:
                    sql('BEGIN IMMEDIATE')
                    begin = True
                    self._txn_id = tid
                    break
                except sqlite3.OperationalError:
                    if retry:
                        continue
                    if filename is not None:
                        _disk_remove(filename)
                    raise Timeout from None

        try:
            yield sql, filenames.append
        except BaseException:
            if begin:
                assert self._txn_id == tid
                self._txn_id = None
                sql('ROLLBACK')
            raise
        else:
            if begin:
                assert self._txn_id == tid
                self._txn_id = None
                sql('COMMIT')
            for name in filenames:
                if name is not None:
                    _disk_remove(name)

    def set(self, key, value, expire=None, read=False, tag=None, retry=False):
        """Set `key` and `value` item in cache.
        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param key: key for item
        :param value: value for item
        :param float expire: seconds until item expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was set
        :raises Timeout: if database timeout occurs
        """
        now = time.time()
        db_key, raw = self._disk.put(key)
        expire_time = None if expire is None else now + expire
        size, mode, filename, db_value = self._disk.store(value, read, key=key)
        columns = (expire_time, tag, size, mode, filename, db_value)

        # The order of SELECT, UPDATE, and INSERT is important below.
        #
        # Typical cache usage pattern is:
        #
        # value = cache.get(key)
        # if value is None:
        #     value = expensive_calculation()
        #     cache.set(key, value)
        #
        # Cache.get does not evict expired keys to avoid writes during lookups.
        # Commonly used/expired keys will therefore remain in the cache making
        # an UPDATE the preferred path.
        #
        # The alternative is to assume the key is not present by first trying
        # to INSERT and then handling the IntegrityError that occurs from
        # violating the UNIQUE constraint. This optimistic approach was
        # rejected based on the common cache usage pattern.
        #
        # INSERT OR REPLACE aka UPSERT is not used because the old filename may
        # need cleanup.

        with self._transact(retry, filename) as (sql, cleanup):
            rows = sql(
                'SELECT rowid, filename FROM Cache'
                ' WHERE key = ? AND raw = ?',
                (db_key, raw),
            ).fetchall()

            if rows:
                ((rowid, old_filename),) = rows
                cleanup(old_filename)
                self._row_update(rowid, now, columns)
            else:
                self._row_insert(db_key, raw, now, columns)

            self._cull(now, sql, cleanup)

            return True

    def __setitem__(self, key, value):
        """Set corresponding `value` for `key` in cache.
        :param key: key for item
        :param value: value for item
        :return: corresponding value
        :raises KeyError: if key is not found
        """
        self.set(key, value, retry=True)

    def _row_update(self, rowid, now, columns):
        sql = self._sql
        expire_time, tag, size, mode, filename, value = columns
        sql(
            'UPDATE Cache SET'
            ' store_time = ?,'
            ' expire_time = ?,'
            ' access_time = ?,'
            ' access_count = ?,'
            ' tag = ?,'
            ' size = ?,'
            ' mode = ?,'
            ' filename = ?,'
            ' value = ?'
            ' WHERE rowid = ?',
            (
                now,  # store_time
                expire_time,
                now,  # access_time
                0,  # access_count
                tag,
                size,
                mode,
                filename,
                value,
                rowid,
            ),
        )

    def _row_insert(self, key, raw, now, columns):
        sql = self._sql
        expire_time, tag, size, mode, filename, value = columns
        sql(
            'INSERT INTO Cache('
            ' key, raw, store_time, expire_time, access_time,'
            ' access_count, tag, size, mode, filename, value'
            ') VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)',
            (
                key,
                raw,
                now,  # store_time
                expire_time,
                now,  # access_time
                0,  # access_count
                tag,
                size,
                mode,
                filename,
                value,
            ),
        )

    def _cull(self, now, sql, cleanup, limit=None):
        cull_limit = self.cull_limit if limit is None else limit

        if cull_limit == 0:
            return

        # Evict expired keys.

        select_expired_template = (
            'SELECT %s FROM Cache'
            ' WHERE expire_time IS NOT NULL AND expire_time < ?'
            ' ORDER BY expire_time LIMIT ?'
        )

        select_expired = select_expired_template % 'filename'
        rows = sql(select_expired, (now, cull_limit)).fetchall()

        if rows:
            delete_expired = 'DELETE FROM Cache WHERE rowid IN (%s)' % (
                select_expired_template % 'rowid'
            )
            sql(delete_expired, (now, cull_limit))

            for (filename,) in rows:
                cleanup(filename)

            cull_limit -= len(rows)

            if cull_limit == 0:
                return

        # Evict keys by policy.

        select_policy = EVICTION_POLICY[self.eviction_policy]['cull']

        if select_policy is None or self.volume() < self.size_limit:
            return

        select_filename = select_policy.format(fields='filename', now=now)
        rows = sql(select_filename, (cull_limit,)).fetchall()

        if rows:
            delete = 'DELETE FROM Cache WHERE rowid IN (%s)' % (
                select_policy.format(fields='rowid', now=now)
            )
            sql(delete, (cull_limit,))

            for (filename,) in rows:
                cleanup(filename)

    def touch(self, key, expire=None, retry=False):
        """Touch `key` in cache and update `expire` time.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param key: key for item
        :param float expire: seconds until item expires
            (default None, no expiry)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if key was touched
        :raises Timeout: if database timeout occurs
        """
        now = time.time()
        db_key, raw = self._disk.put(key)
        expire_time = None if expire is None else now + expire

        with self._transact(retry) as (sql, _):
            rows = sql(
                'SELECT rowid, expire_time FROM Cache'
                ' WHERE key = ? AND raw = ?',
                (db_key, raw),
            ).fetchall()

            if rows:
                ((rowid, old_expire_time),) = rows

                if old_expire_time is None or old_expire_time > now:
                    sql(
                        'UPDATE Cache SET expire_time = ? WHERE rowid = ?',
                        (expire_time, rowid),
                    )
                    return True

        return False

    def add(self, key, value, expire=None, read=False, tag=None, retry=False):
        """Add `key` and `value` item to cache.
        Similar to `set`, but only add to cache if key not present.
        Operation is atomic. Only one concurrent add operation for a given key
        will succeed.
        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param key: key for item
        :param value: value for item
        :param float expire: seconds until the key expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was added
        :raises Timeout: if database timeout occurs
        """
        now = time.time()
        db_key, raw = self._disk.put(key)
        expire_time = None if expire is None else now + expire
        size, mode, filename, db_value = self._disk.store(value, read, key=key)
        columns = (expire_time, tag, size, mode, filename, db_value)

        with self._transact(retry, filename) as (sql, cleanup):
            rows = sql(
                'SELECT rowid, filename, expire_time FROM Cache'
                ' WHERE key = ? AND raw = ?',
                (db_key, raw),
            ).fetchall()

            if rows:
                ((rowid, old_filename, old_expire_time),) = rows

                if old_expire_time is None or old_expire_time > now:
                    cleanup(filename)
                    return False

                cleanup(old_filename)
                self._row_update(rowid, now, columns)
            else:
                self._row_insert(db_key, raw, now, columns)

            self._cull(now, sql, cleanup)

            return True

    def incr(self, key, delta=1, default=0, retry=False):
        """Increment value by delta for item with key.
        If key is missing and default is None then raise KeyError. Else if key
        is missing and default is not None then use default for value.
        Operation is atomic. All concurrent increment operations will be
        counted individually.
        Assumes value may be stored in a SQLite column. Most builds that target
        machines with 64-bit pointer widths will support 64-bit signed
        integers.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param key: key for item
        :param int delta: amount to increment (default 1)
        :param int default: value if key is missing (default 0)
        :param bool retry: retry if database timeout occurs (default False)
        :return: new value for item
        :raises KeyError: if key is not found and default is None
        :raises Timeout: if database timeout occurs
        """
        now = time.time()
        db_key, raw = self._disk.put(key)
        select = (
            'SELECT rowid, expire_time, filename, value FROM Cache'
            ' WHERE key = ? AND raw = ?'
        )

        with self._transact(retry) as (sql, cleanup):
            rows = sql(select, (db_key, raw)).fetchall()

            if not rows:
                if default is None:
                    raise KeyError(key)

                value = default + delta
                columns = (None, None) + self._disk.store(
                    value, False, key=key
                )
                self._row_insert(db_key, raw, now, columns)
                self._cull(now, sql, cleanup)
                return value

            ((rowid, expire_time, filename, value),) = rows

            if expire_time is not None and expire_time < now:
                if default is None:
                    raise KeyError(key)

                value = default + delta
                columns = (None, None) + self._disk.store(
                    value, False, key=key
                )
                self._row_update(rowid, now, columns)
                self._cull(now, sql, cleanup)
                cleanup(filename)
                return value

            value += delta

            columns = 'store_time = ?, value = ?'
            update_column = EVICTION_POLICY[self.eviction_policy]['get']

            if update_column is not None:
                columns += ', ' + update_column.format(now=now)

            update = 'UPDATE Cache SET %s WHERE rowid = ?' % columns
            sql(update, (now, value, rowid))

            return value

    def decr(self, key, delta=1, default=0, retry=False):
        """Decrement value by delta for item with key.
        If key is missing and default is None then raise KeyError. Else if key
        is missing and default is not None then use default for value.
        Operation is atomic. All concurrent decrement operations will be
        counted individually.
        Unlike Memcached, negative values are supported. Value may be
        decremented below zero.
        Assumes value may be stored in a SQLite column. Most builds that target
        machines with 64-bit pointer widths will support 64-bit signed
        integers.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param key: key for item
        :param int delta: amount to decrement (default 1)
        :param int default: value if key is missing (default 0)
        :param bool retry: retry if database timeout occurs (default False)
        :return: new value for item
        :raises KeyError: if key is not found and default is None
        :raises Timeout: if database timeout occurs
        """
        return self.incr(key, -delta, default, retry)

    def get(
        self,
        key,
        default=None,
        read=False,
        expire_time=False,
        tag=False,
        retry=False,
    ):
        """Retrieve value from cache. If `key` is missing, return `default`.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param key: key for item
        :param default: value to return if key is missing (default None)
        :param bool read: if True, return file handle to value
            (default False)
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: value for item or default if key not found
        :raises Timeout: if database timeout occurs
        """
        db_key, raw = self._disk.put(key)
        update_column = EVICTION_POLICY[self.eviction_policy]['get']
        select = (
            'SELECT rowid, expire_time, tag, mode, filename, value'
            ' FROM Cache WHERE key = ? AND raw = ?'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        )

        if expire_time and tag:
            default = (default, None, None)
        elif expire_time or tag:
            default = (default, None)

        if not self.statistics and update_column is None:
            # Fast path, no transaction necessary.

            rows = self._sql(select, (db_key, raw, time.time())).fetchall()

            if not rows:
                return default

            ((rowid, db_expire_time, db_tag, mode, filename, db_value),) = rows

            try:
                value = self._disk.fetch(mode, filename, db_value, read)
            except IOError:
                # Key was deleted before we could retrieve result.
                return default

        else:  # Slow path, transaction required.
            cache_hit = (
                'UPDATE Settings SET value = value + 1 WHERE key = "hits"'
            )
            cache_miss = (
                'UPDATE Settings SET value = value + 1 WHERE key = "misses"'
            )

            with self._transact(retry) as (sql, _):
                rows = sql(select, (db_key, raw, time.time())).fetchall()

                if not rows:
                    if self.statistics:
                        sql(cache_miss)
                    return default

                (
                    (rowid, db_expire_time, db_tag, mode, filename, db_value),
                ) = rows  # noqa: E127

                try:
                    value = self._disk.fetch(mode, filename, db_value, read)
                except IOError as error:
                    if error.errno == errno.ENOENT:
                        # Key was deleted before we could retrieve result.
                        if self.statistics:
                            sql(cache_miss)
                        return default
                    else:
                        raise

                if self.statistics:
                    sql(cache_hit)

                now = time.time()
                update = 'UPDATE Cache SET %s WHERE rowid = ?'

                if update_column is not None:
                    sql(update % update_column.format(now=now), (rowid,))

        if expire_time and tag:
            return (value, db_expire_time, db_tag)
        elif expire_time:
            return (value, db_expire_time)
        elif tag:
            return (value, db_tag)
        else:
            return value

    def __getitem__(self, key):
        """Return corresponding value for `key` from cache.
        :param key: key matching item
        :return: corresponding value
        :raises KeyError: if key is not found
        """
        value = self.get(key, default=ENOVAL, retry=True)
        if value is ENOVAL:
            raise KeyError(key)
        return value

    def read(self, key, retry=False):
        """Return file handle value corresponding to `key` from cache.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param key: key matching item
        :param bool retry: retry if database timeout occurs (default False)
        :return: file open for reading in binary mode
        :raises KeyError: if key is not found
        :raises Timeout: if database timeout occurs
        """
        handle = self.get(key, default=ENOVAL, read=True, retry=retry)
        if handle is ENOVAL:
            raise KeyError(key)
        return handle

    def __contains__(self, key):
        """Return `True` if `key` matching item is found in cache.
        :param key: key matching item
        :return: True if key matching item
        """
        sql = self._sql
        db_key, raw = self._disk.put(key)
        select = (
            'SELECT rowid FROM Cache'
            ' WHERE key = ? AND raw = ?'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        )

        rows = sql(select, (db_key, raw, time.time())).fetchall()

        return bool(rows)

    def pop(
        self, key, default=None, expire_time=False, tag=False, retry=False
    ):  # noqa: E501
        """Remove corresponding item for `key` from cache and return value.
        If `key` is missing, return `default`.
        Operation is atomic. Concurrent operations will be serialized.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param key: key for item
        :param default: value to return if key is missing (default None)
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: value for item or default if key not found
        :raises Timeout: if database timeout occurs
        """
        db_key, raw = self._disk.put(key)
        select = (
            'SELECT rowid, expire_time, tag, mode, filename, value'
            ' FROM Cache WHERE key = ? AND raw = ?'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        )

        if expire_time and tag:
            default = default, None, None
        elif expire_time or tag:
            default = default, None

        with self._transact(retry) as (sql, _):
            rows = sql(select, (db_key, raw, time.time())).fetchall()

            if not rows:
                return default

            ((rowid, db_expire_time, db_tag, mode, filename, db_value),) = rows

            sql('DELETE FROM Cache WHERE rowid = ?', (rowid,))

        try:
            value = self._disk.fetch(mode, filename, db_value, False)
        except IOError as error:
            if error.errno == errno.ENOENT:
                # Key was deleted before we could retrieve result.
                return default
            else:
                raise
        finally:
            if filename is not None:
                self._disk.remove(filename)

        if expire_time and tag:
            return value, db_expire_time, db_tag
        elif expire_time:
            return value, db_expire_time
        elif tag:
            return value, db_tag
        else:
            return value

    def __delitem__(self, key, retry=True):
        """Delete corresponding item for `key` from cache.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default `True`).
        :param key: key matching item
        :param bool retry: retry if database timeout occurs (default True)
        :raises KeyError: if key is not found
        :raises Timeout: if database timeout occurs
        """
        db_key, raw = self._disk.put(key)

        with self._transact(retry) as (sql, cleanup):
            rows = sql(
                'SELECT rowid, filename FROM Cache'
                ' WHERE key = ? AND raw = ?'
                ' AND (expire_time IS NULL OR expire_time > ?)',
                (db_key, raw, time.time()),
            ).fetchall()

            if not rows:
                raise KeyError(key)

            ((rowid, filename),) = rows
            sql('DELETE FROM Cache WHERE rowid = ?', (rowid,))
            cleanup(filename)

            return True

    def delete(self, key, retry=False):
        """Delete corresponding item for `key` from cache.
        Missing keys are ignored.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param key: key matching item
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was deleted
        :raises Timeout: if database timeout occurs
        """
        try:
            return self.__delitem__(key, retry=retry)
        except KeyError:
            return False

    def push(
        self,
        value,
        prefix=None,
        side='back',
        expire=None,
        read=False,
        tag=None,
        retry=False,
    ):
        """Push `value` onto `side` of queue identified by `prefix` in cache.
        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.
        Defaults to pushing value on back of queue. Set side to 'front' to push
        value on front of queue. Side must be one of 'back' or 'front'.
        Operation is atomic. Concurrent operations will be serialized.
        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        See also `Cache.pull`.
        >>> cache = Cache()
        >>> print(cache.push('first value'))
        500000000000000
        >>> cache.get(500000000000000)
        'first value'
        >>> print(cache.push('second value'))
        500000000000001
        >>> print(cache.push('third value', side='front'))
        499999999999999
        >>> cache.push(1234, prefix='userids')
        'userids-500000000000000'
        :param value: value for item
        :param str prefix: key prefix (default None, key is integer)
        :param str side: either 'back' or 'front' (default 'back')
        :param float expire: seconds until the key expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key for item in cache
        :raises Timeout: if database timeout occurs
        """
        if prefix is None:
            min_key = 0
            max_key = 999999999999999
        else:
            min_key = prefix + '-000000000000000'
            max_key = prefix + '-999999999999999'

        now = time.time()
        raw = True
        expire_time = None if expire is None else now + expire
        size, mode, filename, db_value = self._disk.store(value, read)
        columns = (expire_time, tag, size, mode, filename, db_value)
        order = {'back': 'DESC', 'front': 'ASC'}
        select = (
            'SELECT key FROM Cache'
            ' WHERE ? < key AND key < ? AND raw = ?'
            ' ORDER BY key %s LIMIT 1'
        ) % order[side]

        with self._transact(retry, filename) as (sql, cleanup):
            rows = sql(select, (min_key, max_key, raw)).fetchall()

            if rows:
                ((key,),) = rows

                if prefix is not None:
                    num = int(key[(key.rfind('-') + 1) :])
                else:
                    num = key

                if side == 'back':
                    num += 1
                else:
                    assert side == 'front'
                    num -= 1
            else:
                num = 500000000000000

            if prefix is not None:
                db_key = '{0}-{1:015d}'.format(prefix, num)
            else:
                db_key = num

            self._row_insert(db_key, raw, now, columns)
            self._cull(now, sql, cleanup)

            return db_key

    def pull(
        self,
        prefix=None,
        default=(None, None),
        side='front',
        expire_time=False,
        tag=False,
        retry=False,
    ):
        """Pull key and value item pair from `side` of queue in cache.
        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.
        If queue is empty, return default.
        Defaults to pulling key and value item pairs from front of queue. Set
        side to 'back' to pull from back of queue. Side must be one of 'front'
        or 'back'.
        Operation is atomic. Concurrent operations will be serialized.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        See also `Cache.push` and `Cache.get`.
        >>> cache = Cache()
        >>> cache.pull()
        (None, None)
        >>> for letter in 'abc':
        ...     print(cache.push(letter))
        500000000000000
        500000000000001
        500000000000002
        >>> key, value = cache.pull()
        >>> print(key)
        500000000000000
        >>> value
        'a'
        >>> _, value = cache.pull(side='back')
        >>> value
        'c'
        >>> cache.push(1234, 'userids')
        'userids-500000000000000'
        >>> _, value = cache.pull('userids')
        >>> value
        1234
        :param str prefix: key prefix (default None, key is integer)
        :param default: value to return if key is missing
            (default (None, None))
        :param str side: either 'front' or 'back' (default 'front')
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key and value item pair or default if queue is empty
        :raises Timeout: if database timeout occurs
        """
        # Caution: Nearly identical code exists in Cache.peek
        if prefix is None:
            min_key = 0
            max_key = 999999999999999
        else:
            min_key = prefix + '-000000000000000'
            max_key = prefix + '-999999999999999'

        order = {'front': 'ASC', 'back': 'DESC'}
        select = (
            'SELECT rowid, key, expire_time, tag, mode, filename, value'
            ' FROM Cache WHERE ? < key AND key < ? AND raw = 1'
            ' ORDER BY key %s LIMIT 1'
        ) % order[side]

        if expire_time and tag:
            default = default, None, None
        elif expire_time or tag:
            default = default, None

        while True:
            while True:
                with self._transact(retry) as (sql, cleanup):
                    rows = sql(select, (min_key, max_key)).fetchall()

                    if not rows:
                        return default

                    (
                        (rowid, key, db_expire, db_tag, mode, name, db_value),
                    ) = rows

                    sql('DELETE FROM Cache WHERE rowid = ?', (rowid,))

                    if db_expire is not None and db_expire < time.time():
                        cleanup(name)
                    else:
                        break

            try:
                value = self._disk.fetch(mode, name, db_value, False)
            except IOError as error:
                if error.errno == errno.ENOENT:
                    # Key was deleted before we could retrieve result.
                    continue
                raise
            finally:
                if name is not None:
                    self._disk.remove(name)
            break

        if expire_time and tag:
            return (key, value), db_expire, db_tag
        elif expire_time:
            return (key, value), db_expire
        elif tag:
            return (key, value), db_tag
        else:
            return key, value

    def peek(
        self,
        prefix=None,
        default=(None, None),
        side='front',
        expire_time=False,
        tag=False,
        retry=False,
    ):
        """Peek at key and value item pair from `side` of queue in cache.
        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.
        If queue is empty, return default.
        Defaults to peeking at key and value item pairs from front of queue.
        Set side to 'back' to pull from back of queue. Side must be one of
        'front' or 'back'.
        Expired items are deleted from cache. Operation is atomic. Concurrent
        operations will be serialized.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        See also `Cache.pull` and `Cache.push`.
        >>> cache = Cache()
        >>> for letter in 'abc':
        ...     print(cache.push(letter))
        500000000000000
        500000000000001
        500000000000002
        >>> key, value = cache.peek()
        >>> print(key)
        500000000000000
        >>> value
        'a'
        >>> key, value = cache.peek(side='back')
        >>> print(key)
        500000000000002
        >>> value
        'c'
        :param str prefix: key prefix (default None, key is integer)
        :param default: value to return if key is missing
            (default (None, None))
        :param str side: either 'front' or 'back' (default 'front')
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key and value item pair or default if queue is empty
        :raises Timeout: if database timeout occurs
        """
        # Caution: Nearly identical code exists in Cache.pull
        if prefix is None:
            min_key = 0
            max_key = 999999999999999
        else:
            min_key = prefix + '-000000000000000'
            max_key = prefix + '-999999999999999'

        order = {'front': 'ASC', 'back': 'DESC'}
        select = (
            'SELECT rowid, key, expire_time, tag, mode, filename, value'
            ' FROM Cache WHERE ? < key AND key < ? AND raw = 1'
            ' ORDER BY key %s LIMIT 1'
        ) % order[side]

        if expire_time and tag:
            default = default, None, None
        elif expire_time or tag:
            default = default, None

        while True:
            while True:
                with self._transact(retry) as (sql, cleanup):
                    rows = sql(select, (min_key, max_key)).fetchall()

                    if not rows:
                        return default

                    (
                        (rowid, key, db_expire, db_tag, mode, name, db_value),
                    ) = rows

                    if db_expire is not None and db_expire < time.time():
                        sql('DELETE FROM Cache WHERE rowid = ?', (rowid,))
                        cleanup(name)
                    else:
                        break

            try:
                value = self._disk.fetch(mode, name, db_value, False)
            except IOError as error:
                if error.errno == errno.ENOENT:
                    # Key was deleted before we could retrieve result.
                    continue
                raise
            finally:
                if name is not None:
                    self._disk.remove(name)
            break

        if expire_time and tag:
            return (key, value), db_expire, db_tag
        elif expire_time:
            return (key, value), db_expire
        elif tag:
            return (key, value), db_tag
        else:
            return key, value

    def peekitem(self, last=True, expire_time=False, tag=False, retry=False):
        """Peek at key and value item pair in cache based on iteration order.
        Expired items are deleted from cache. Operation is atomic. Concurrent
        operations will be serialized.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        >>> cache = Cache()
        >>> for num, letter in enumerate('abc'):
        ...     cache[letter] = num
        >>> cache.peekitem()
        ('c', 2)
        >>> cache.peekitem(last=False)
        ('a', 0)
        :param bool last: last item in iteration order (default True)
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key and value item pair
        :raises KeyError: if cache is empty
        :raises Timeout: if database timeout occurs
        """
        order = ('ASC', 'DESC')
        select = (
            'SELECT rowid, key, raw, expire_time, tag, mode, filename, value'
            ' FROM Cache ORDER BY rowid %s LIMIT 1'
        ) % order[last]

        while True:
            while True:
                with self._transact(retry) as (sql, cleanup):
                    rows = sql(select).fetchall()

                    if not rows:
                        raise KeyError('dictionary is empty')

                    (
                        (
                            rowid,
                            db_key,
                            raw,
                            db_expire,
                            db_tag,
                            mode,
                            name,
                            db_value,
                        ),
                    ) = rows

                    if db_expire is not None and db_expire < time.time():
                        sql('DELETE FROM Cache WHERE rowid = ?', (rowid,))
                        cleanup(name)
                    else:
                        break

            key = self._disk.get(db_key, raw)

            try:
                value = self._disk.fetch(mode, name, db_value, False)
            except IOError as error:
                if error.errno == errno.ENOENT:
                    # Key was deleted before we could retrieve result.
                    continue
                raise
            break

        if expire_time and tag:
            return (key, value), db_expire, db_tag
        elif expire_time:
            return (key, value), db_expire
        elif tag:
            return (key, value), db_tag
        else:
            return key, value

    def memoize(self, name=None, typed=False, expire=None, tag=None):
        """Memoizing cache decorator.
        Decorator to wrap callable with memoizing function using cache.
        Repeated calls with the same arguments will lookup result in cache and
        avoid function evaluation.
        If name is set to None (default), the callable name will be determined
        automatically.
        When expire is set to zero, function results will not be set in the
        cache. Cache lookups still occur, however. Read
        :doc:`case-study-landing-page-caching` for example usage.
        If typed is set to True, function arguments of different types will be
        cached separately. For example, f(3) and f(3.0) will be treated as
        distinct calls with distinct results.
        The original underlying function is accessible through the __wrapped__
        attribute. This is useful for introspection, for bypassing the cache,
        or for rewrapping the function with a different cache.
        >>> from diskcache import Cache
        >>> cache = Cache()
        >>> @cache.memoize(expire=1, tag='fib')
        ... def fibonacci(number):
        ...     if number == 0:
        ...         return 0
        ...     elif number == 1:
        ...         return 1
        ...     else:
        ...         return fibonacci(number - 1) + fibonacci(number - 2)
        >>> print(fibonacci(100))
        354224848179261915075
        An additional `__cache_key__` attribute can be used to generate the
        cache key used for the given arguments.
        >>> key = fibonacci.__cache_key__(100)
        >>> print(cache[key])
        354224848179261915075
        Remember to call memoize when decorating a callable. If you forget,
        then a TypeError will occur. Note the lack of parenthenses after
        memoize below:
        >>> @cache.memoize
        ... def test():
        ...     pass
        Traceback (most recent call last):
            ...
        TypeError: name cannot be callable
        :param cache: cache to store callable arguments and return values
        :param str name: name given for callable (default None, automatic)
        :param bool typed: cache different types separately (default False)
        :param float expire: seconds until arguments expire
            (default None, no expiry)
        :param str tag: text to associate with arguments (default None)
        :return: callable decorator
        """
        # Caution: Nearly identical code exists in DjangoCache.memoize
        if callable(name):
            raise TypeError('name cannot be callable')

        def decorator(func):
            "Decorator created by memoize() for callable `func`."
            base = (full_name(func),) if name is None else (name,)

            if not inspect.iscoroutinefunction(func):
                @ft.wraps(func)
                def wrapper(*args, **kwargs):
                    "Wrapper for callable to cache arguments and return values."
                    key = wrapper.__cache_key__(*args, **kwargs)
                    result = self.get(key, default=ENOVAL, retry=True)

                    if result is ENOVAL:
                        result = func(*args, **kwargs)
                        if expire is None or expire > 0:
                            self.set(key, result, expire, tag=tag, retry=True)

                    return result
            else:
                @ft.wraps(func)
                async def wrapper(*args, **kwargs):
                    "Wrapper for callable to cache arguments and return values."
                    key = wrapper.__cache_key__(*args, **kwargs)
                    result = self.get(key, default=ENOVAL, retry=True)

                    if result is ENOVAL:
                        result = await func(*args, **kwargs)
                        if expire is None or expire > 0:
                            self.set(key, result, expire, tag=tag, retry=True)

                    return result

            def __cache_key__(*args, **kwargs):
                "Make key for cache given function arguments."
                return args_to_key(base, args, kwargs, typed)

            wrapper.__cache_key__ = __cache_key__
            return wrapper

        return decorator

    def check(self, fix=False, retry=False):
        """Check database and file system consistency.
        Intended for use in testing and post-mortem error analysis.
        While checking the Cache table for consistency, a writer lock is held
        on the database. The lock blocks other cache clients from writing to
        the database. For caches with many file references, the lock may be
        held for a long time. For example, local benchmarking shows that a
        cache with 1,000 file references takes ~60ms to check.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param bool fix: correct inconsistencies
        :param bool retry: retry if database timeout occurs (default False)
        :return: list of warnings
        :raises Timeout: if database timeout occurs
        """
        # pylint: disable=access-member-before-definition,W0201
        with warnings.catch_warnings(record=True) as warns:
            sql = self._sql

            # Check integrity of database.

            rows = sql('PRAGMA integrity_check').fetchall()

            if len(rows) != 1 or rows[0][0] != u'ok':
                for (message,) in rows:
                    warnings.warn(message)

            if fix:
                sql('VACUUM')

            with self._transact(retry) as (sql, _):

                # Check Cache.filename against file system.

                filenames = set()
                select = (
                    'SELECT rowid, size, filename FROM Cache'
                    ' WHERE filename IS NOT NULL'
                )

                rows = sql(select).fetchall()

                for rowid, size, filename in rows:
                    full_path = op.join(self._directory, filename)
                    filenames.add(full_path)

                    if op.exists(full_path):
                        real_size = op.getsize(full_path)

                        if size != real_size:
                            message = 'wrong file size: %s, %d != %d'
                            args = full_path, real_size, size
                            warnings.warn(message % args)

                            if fix:
                                sql(
                                    'UPDATE Cache SET size = ?'
                                    ' WHERE rowid = ?',
                                    (real_size, rowid),
                                )

                        continue

                    warnings.warn('file not found: %s' % full_path)

                    if fix:
                        sql('DELETE FROM Cache WHERE rowid = ?', (rowid,))

                # Check file system against Cache.filename.

                for dirpath, _, files in os.walk(self._directory):
                    paths = [op.join(dirpath, filename) for filename in files]
                    error = set(paths) - filenames

                    for full_path in error:
                        if DBNAME in full_path:
                            continue

                        message = 'unknown file: %s' % full_path
                        warnings.warn(message, UnknownFileWarning)

                        if fix:
                            os.remove(full_path)

                # Check for empty directories.

                for dirpath, dirs, files in os.walk(self._directory):
                    if not (dirs or files):
                        message = 'empty directory: %s' % dirpath
                        warnings.warn(message, EmptyDirWarning)

                        if fix:
                            os.rmdir(dirpath)

                # Check Settings.count against count of Cache rows.

                self.reset('count')
                ((count,),) = sql('SELECT COUNT(key) FROM Cache').fetchall()

                if self.count != count:
                    message = 'Settings.count != COUNT(Cache.key); %d != %d'
                    warnings.warn(message % (self.count, count))

                    if fix:
                        sql(
                            'UPDATE Settings SET value = ? WHERE key = ?',
                            (count, 'count'),
                        )

                # Check Settings.size against sum of Cache.size column.

                self.reset('size')
                select_size = 'SELECT COALESCE(SUM(size), 0) FROM Cache'
                ((size,),) = sql(select_size).fetchall()

                if self.size != size:
                    message = 'Settings.size != SUM(Cache.size); %d != %d'
                    warnings.warn(message % (self.size, size))

                    if fix:
                        sql(
                            'UPDATE Settings SET value = ? WHERE key =?',
                            (size, 'size'),
                        )

            return warns

    def create_tag_index(self):
        """Create tag index on cache database.
        It is better to initialize cache with `tag_index=True` than use this.
        :raises Timeout: if database timeout occurs
        """
        sql = self._sql
        sql('CREATE INDEX IF NOT EXISTS Cache_tag_rowid ON Cache(tag, rowid)')
        self.reset('tag_index', 1)

    def drop_tag_index(self):
        """Drop tag index on cache database.
        :raises Timeout: if database timeout occurs
        """
        sql = self._sql
        sql('DROP INDEX IF EXISTS Cache_tag_rowid')
        self.reset('tag_index', 0)

    def evict(self, tag, retry=False):
        """Remove items with matching `tag` from cache.
        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.
        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param str tag: tag identifying items
        :param bool retry: retry if database timeout occurs (default False)
        :return: count of rows removed
        :raises Timeout: if database timeout occurs
        """
        select = (
            'SELECT rowid, filename FROM Cache'
            ' WHERE tag = ? AND rowid > ?'
            ' ORDER BY rowid LIMIT ?'
        )
        args = [tag, 0, 100]
        return self._select_delete(select, args, arg_index=1, retry=retry)

    def expire(self, now=None, retry=False):
        """Remove expired items from cache.
        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.
        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param float now: current time (default None, ``time.time()`` used)
        :param bool retry: retry if database timeout occurs (default False)
        :return: count of items removed
        :raises Timeout: if database timeout occurs
        """
        select = (
            'SELECT rowid, expire_time, filename FROM Cache'
            ' WHERE ? < expire_time AND expire_time < ?'
            ' ORDER BY expire_time LIMIT ?'
        )
        args = [0, now or time.time(), 100]
        return self._select_delete(select, args, row_index=1, retry=retry)

    def cull(self, retry=False):
        """Cull items from cache until volume is less than size limit.
        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.
        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param bool retry: retry if database timeout occurs (default False)
        :return: count of items removed
        :raises Timeout: if database timeout occurs
        """
        now = time.time()

        # Remove expired items.

        count = self.expire(now)

        # Remove items by policy.

        select_policy = EVICTION_POLICY[self.eviction_policy]['cull']

        if select_policy is None:
            return 0

        select_filename = select_policy.format(fields='filename', now=now)

        try:
            while self.volume() > self.size_limit:
                with self._transact(retry) as (sql, cleanup):
                    rows = sql(select_filename, (10,)).fetchall()

                    if not rows:
                        break

                    count += len(rows)
                    delete = (
                        'DELETE FROM Cache WHERE rowid IN (%s)'
                        % select_policy.format(fields='rowid', now=now)
                    )
                    sql(delete, (10,))

                    for (filename,) in rows:
                        cleanup(filename)
        except Timeout:
            raise Timeout(count) from None

        return count

    def clear(self, retry=False):
        """Remove all items from cache.
        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.
        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.
        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param bool retry: retry if database timeout occurs (default False)
        :return: count of rows removed
        :raises Timeout: if database timeout occurs
        """
        select = (
            'SELECT rowid, filename FROM Cache'
            ' WHERE rowid > ?'
            ' ORDER BY rowid LIMIT ?'
        )
        args = [0, 100]
        return self._select_delete(select, args, retry=retry)

    def _select_delete(
        self, select, args, row_index=0, arg_index=0, retry=False
    ):
        count = 0
        delete = 'DELETE FROM Cache WHERE rowid IN (%s)'

        try:
            while True:
                with self._transact(retry) as (sql, cleanup):
                    rows = sql(select, args).fetchall()

                    if not rows:
                        break

                    count += len(rows)
                    sql(delete % ','.join(str(row[0]) for row in rows))

                    for row in rows:
                        args[arg_index] = row[row_index]
                        cleanup(row[-1])

        except Timeout:
            raise Timeout(count) from None

        return count

    def iterkeys(self, reverse=False):
        """Iterate Cache keys in database sort order.
        >>> cache = Cache()
        >>> for key in [4, 1, 3, 0, 2]:
        ...     cache[key] = key
        >>> list(cache.iterkeys())
        [0, 1, 2, 3, 4]
        >>> list(cache.iterkeys(reverse=True))
        [4, 3, 2, 1, 0]
        :param bool reverse: reverse sort order (default False)
        :return: iterator of Cache keys
        """
        sql = self._sql
        limit = 100
        _disk_get = self._disk.get

        if reverse:
            select = (
                'SELECT key, raw FROM Cache'
                ' ORDER BY key DESC, raw DESC LIMIT 1'
            )
            iterate = (
                'SELECT key, raw FROM Cache'
                ' WHERE key = ? AND raw < ? OR key < ?'
                ' ORDER BY key DESC, raw DESC LIMIT ?'
            )
        else:
            select = (
                'SELECT key, raw FROM Cache'
                ' ORDER BY key ASC, raw ASC LIMIT 1'
            )
            iterate = (
                'SELECT key, raw FROM Cache'
                ' WHERE key = ? AND raw > ? OR key > ?'
                ' ORDER BY key ASC, raw ASC LIMIT ?'
            )

        row = sql(select).fetchall()

        if row:
            ((key, raw),) = row
        else:
            return

        yield _disk_get(key, raw)

        while True:
            rows = sql(iterate, (key, raw, key, limit)).fetchall()

            if not rows:
                break

            for key, raw in rows:
                yield _disk_get(key, raw)

    def _iter(self, ascending=True):
        sql = self._sql
        rows = sql('SELECT MAX(rowid) FROM Cache').fetchall()
        ((max_rowid,),) = rows
        yield  # Signal ready.

        if max_rowid is None:
            return

        bound = max_rowid + 1
        limit = 100
        _disk_get = self._disk.get
        rowid = 0 if ascending else bound
        select = (
            'SELECT rowid, key, raw FROM Cache'
            ' WHERE ? < rowid AND rowid < ?'
            ' ORDER BY rowid %s LIMIT ?'
        ) % ('ASC' if ascending else 'DESC')

        while True:
            if ascending:
                args = (rowid, bound, limit)
            else:
                args = (0, rowid, limit)

            rows = sql(select, args).fetchall()

            if not rows:
                break

            for rowid, key, raw in rows:
                yield _disk_get(key, raw)

    def __iter__(self):
        "Iterate keys in cache including expired items."
        iterator = self._iter()
        next(iterator)
        return iterator

    def __reversed__(self):
        "Reverse iterate keys in cache including expired items."
        iterator = self._iter(ascending=False)
        next(iterator)
        return iterator

    def stats(self, enable=True, reset=False):
        """Return cache statistics hits and misses.
        :param bool enable: enable collecting statistics (default True)
        :param bool reset: reset hits and misses to 0 (default False)
        :return: (hits, misses)
        """
        # pylint: disable=E0203,W0201
        result = (self.reset('hits'), self.reset('misses'))

        if reset:
            self.reset('hits', 0)
            self.reset('misses', 0)

        self.reset('statistics', enable)

        return result

    def volume(self):
        """Return estimated total size of cache on disk.
        :return: size in bytes
        """
        ((page_count,),) = self._sql('PRAGMA page_count').fetchall()
        total_size = self._page_size * page_count + self.reset('size')
        return total_size

    def close(self):
        """Close database connection."""
        con = getattr(self._local, 'con', None)

        if con is None:
            return

        con.close()

        try:
            delattr(self._local, 'con')
        except AttributeError:
            pass

    def __enter__(self):
        # Create connection in thread.
        # pylint: disable=unused-variable
        connection = self._con  # noqa
        return self

    def __exit__(self, *exception):
        self.close()

    def __len__(self):
        "Count of items in cache including expired items."
        return self.reset('count')

    def __getstate__(self):
        return (self.directory, self.timeout, type(self.disk))

    def __setstate__(self, state):
        self.__init__(*state)

    def reset(self, key, value=ENOVAL, update=True):
        """Reset `key` and `value` item from Settings table.
        Use `reset` to update the value of Cache settings correctly. Cache
        settings are stored in the Settings table of the SQLite database. If
        `update` is ``False`` then no attempt is made to update the database.
        If `value` is not given, it is reloaded from the Settings
        table. Otherwise, the Settings table is updated.
        Settings with the ``disk_`` prefix correspond to Disk
        attributes. Updating the value will change the unprefixed attribute on
        the associated Disk instance.
        Settings with the ``sqlite_`` prefix correspond to SQLite
        pragmas. Updating the value will execute the corresponding PRAGMA
        statement.
        SQLite PRAGMA statements may be executed before the Settings table
        exists in the database by setting `update` to ``False``.
        :param str key: Settings key for item
        :param value: value for item (optional)
        :param bool update: update database Settings table (default True)
        :return: updated value for item
        :raises Timeout: if database timeout occurs
        """
        sql = self._sql
        sql_retry = self._sql_retry

        if value is ENOVAL:
            select = 'SELECT value FROM Settings WHERE key = ?'
            ((value,),) = sql_retry(select, (key,)).fetchall()
            setattr(self, key, value)
            return value

        if update:
            statement = 'UPDATE Settings SET value = ? WHERE key = ?'
            sql_retry(statement, (value, key))

        if key.startswith('sqlite_'):
            pragma = key[7:]

            # 2016-02-17 GrantJ - PRAGMA and isolation_level=None
            # don't always play nicely together. Retry setting the
            # PRAGMA. I think some PRAGMA statements expect to
            # immediately take an EXCLUSIVE lock on the database. I
            # can't find any documentation for this but without the
            # retry, stress will intermittently fail with multiple
            # processes.

            # 2018-11-05 GrantJ - Avoid setting pragma values that
            # are already set. Pragma settings like auto_vacuum and
            # journal_mode can take a long time or may not work after
            # tables have been created.

            start = time.time()
            while True:
                try:
                    try:
                        ((old_value,),) = sql(
                            'PRAGMA %s' % (pragma)
                        ).fetchall()
                        update = old_value != value
                    except ValueError:
                        update = True
                    if update:
                        sql('PRAGMA %s = %s' % (pragma, value)).fetchall()
                    break
                except sqlite3.OperationalError as exc:
                    if str(exc) != 'database is locked':
                        raise
                    diff = time.time() - start
                    if diff > 60:
                        raise
                    time.sleep(0.001)
        elif key.startswith('disk_'):
            attr = key[5:]
            setattr(self._disk, attr, value)

        setattr(self, key, value)
        return value


"""
Serializers that depend on/combine with one another
"""

__all__ = (
    'YamlBase64', 'YamlBGZ', 
    'JsonBase64', 'JasonBGZ'

)

from typing import Dict, Any, List, Union
from .core import Defaults

from ._json import Json
from ._yaml import Yaml
from ._base import Base


class YamlBase64(Yaml):
    
    @classmethod
    def dumps(cls, obj: Dict[Any, Any], dumper: str = Defaults.yaml_dumper, *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        rez = cls._encode(obj, dumper = dumper, *args, default = default, **kwargs)
        return Base.b64_encode(rez, *args, **kwargs)

    @classmethod
    def loads(cls, data: Union[str, bytes], loader: str = Defaults.yaml_loader, *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        rez = Base.b64_decode(data, *args, **kwargs)
        return cls._decode(rez, loader = loader, *args, **kwargs)
    
    @classmethod
    async def async_dumps(cls, obj: Dict[Any, Any], dumper: str = Defaults.yaml_dumper, *args, default: Any = None, **kwargs) -> str:
        if not cls.async_supported: raise Exception
        rez = await cls._async_encode(obj, dumper = dumper, *args, default = default, **kwargs)
        return Base.b64_encode(rez, *args, **kwargs)

    @classmethod
    async def async_loads(cls, data: Union[str, bytes], loader: str = Defaults.yaml_loader, *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if not cls.async_supported: raise Exception
        rez = Base.b64_decode(data, *args, **kwargs)
        return await cls._async_decode(rez, loader = loader, *args, **kwargs)

class YamlBGZ(Yaml):

    @classmethod
    def dumps(cls, obj: Dict[Any, Any], dumper: str = Defaults.yaml_dumper, *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        rez = cls._encode(obj, dumper = dumper, *args, default = default, **kwargs)
        return Base.b64_gzip_encode(rez, *args, **kwargs)

    @classmethod
    def loads(cls, data: Union[str, bytes], loader: str = Defaults.yaml_loader, *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        rez = Base.b64_gzip_decode(data, *args, **kwargs)
        return cls._decode(rez, loader = loader, *args, **kwargs)
    
    @classmethod
    async def async_dumps(cls, obj: Dict[Any, Any], dumper: str = Defaults.yaml_dumper, *args, default: Any = None, **kwargs) -> str:
        if not cls.async_supported: raise Exception
        rez = await cls._async_encode(obj, dumper = dumper, *args, default = default, **kwargs)
        return Base.b64_gzip_encode(rez, *args, **kwargs)

    @classmethod
    async def async_loads(cls, data: Union[str, bytes], loader: str = Defaults.yaml_loader, *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if not cls.async_supported: raise Exception
        rez = Base.b64_gzip_decode(data, *args, **kwargs)
        return await cls._async_decode(rez, loader = loader, *args, **kwargs)

class JsonBase64(Json):

    @classmethod
    def dumps(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        rez = cls._encode(obj, *args, default = default, **kwargs)
        return Base.b64_encode(rez, *args, **kwargs)

    @classmethod
    def loads(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        rez = Base.b64_decode(data, *args, **kwargs)
        return cls._decode(rez, *args, **kwargs)
    
    @classmethod
    async def async_dumps(cls, obj: Dict[Any, Any], *args, default: Any = None, **kwargs) -> str:
        if not cls.async_supported: raise Exception
        rez = await cls._async_encode(obj, *args, default = default, **kwargs)
        return Base.b64_encode(rez, *args, **kwargs)

    @classmethod
    async def async_loads(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if not cls.async_supported: raise Exception
        rez = Base.b64_decode(data, *args, **kwargs)
        return await cls._async_decode(rez, *args, **kwargs)


class JsonBGZ(Json):

    @classmethod
    def dumps(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        rez = cls._encode(obj, *args, default = default, **kwargs)
        return Base.b64_gzip_encode(rez, *args, **kwargs)

    @classmethod
    def loads(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        rez = Base.b64_gzip_decode(data, *args, **kwargs)
        return cls._decode(rez, *args, **kwargs)
    
    @classmethod
    async def async_dumps(cls, obj: Dict[Any, Any], *args, default: Any = None, **kwargs) -> str:
        if not cls.async_supported: raise Exception
        rez = await cls._async_encode(obj, *args, default = default, **kwargs)
        return Base.b64_gzip_encode(rez, *args, **kwargs)

    @classmethod
    async def async_loads(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if not cls.async_supported: raise Exception
        rez = Base.b64_gzip_decode(data, *args, **kwargs)
        return await cls._async_decode(rez, *args, **kwargs)



import orjson as _orjson
import simdjson as _simdjson

from typing import Dict, Any, List, Union, overload
from os import PathLike
from pydantic import BaseModel
from .base import Srlzer, Mode

    
class JsonSrlzer(Srlzer):
    default_value: Dict[Any, Any] = None
    async_supported: bool = True
    cloud_supported: bool = True
    
    @classmethod
    def dumps(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        return cls._encode(obj, *args, default = default, **kwargs)

    @classmethod
    def loads(cls, obj: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        return cls._decode(obj, *args, **kwargs)
    
    @classmethod
    def _encode(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        raise NotImplementedError
        
    @classmethod
    def _decode(cls, obj: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        raise NotImplementedError
    
    @classmethod
    async def async_dumps(cls, obj: Dict[Any, Any], *args, default: Any = None, **kwargs) -> str:
        if not cls.async_supported: raise Exception
        return await cls._async_encode(obj, *args, default = default, **kwargs)

    @classmethod
    async def async_loads(cls, obj: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if not cls.async_supported: raise Exception
        return await cls._async_decode(obj, *args, **kwargs)
    
    @classmethod
    async def _async_encode(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        raise NotImplementedError
    
    @classmethod
    async def _async_decode(cls, obj: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        raise NotImplementedError

    @classmethod
    def loadlines(cls, path: PathLike, mode: str = Mode.read, encoding: str = 'utf-8', as_iterable: bool = False, ignore_errors: bool = True, *args, **kwargs):
        ## Change Later
        from pathlib import Path
        path = Path(path)
        if as_iterable:
            def jsonlines_iterator():
                with path.open(mode=mode, encoding=encoding) as f:
                    for line in f:
                        try: yield cls.loads(line, *args, **kwargs)
                        except StopIteration: break
                        except Exception as e:
                            if not ignore_errors: raise e
            return jsonlines_iterator                
        rez = []
        with path.open(mode=mode, encoding=encoding) as f:
            for line in f:
                try: rez.append(cls.loads(line, *args, **kwargs))
                except Exception as e:
                    if not ignore_errors: raise e
        return rez
    
    @classmethod
    def iterlines(cls, path: PathLike, mode: str = Mode.read, encoding: str = 'utf-8', ignore_errors: bool = True, *args, **kwargs):
        from pathlib import Path
        path = Path(path)
        with path.open(mode=mode, encoding=encoding) as f:
            for line in f:
                try: yield cls.loads(line, *args, **kwargs)
                except StopIteration: break
                except Exception as e:
                    if not ignore_errors: raise e



class OrJson(JsonSrlzer):
    
    @classmethod
    def _encode(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        return _orjson.dumps(obj, default=default, *args, **kwargs).decode()
    
    @classmethod
    def _decode(cls, obj: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        return _orjson.loads(obj, *args, **kwargs)
    
    @classmethod
    async def _async_encode(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        return _orjson.dumps(obj, default=default, *args, **kwargs).decode()
    
    @classmethod
    async def _async_decode(cls, obj: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        return _orjson.loads(obj, *args, **kwargs)


class SimdJson(JsonSrlzer):
    parser: _simdjson.Parser = _simdjson.Parser()
    
    @classmethod
    def _encode(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        return _simdjson.dumps(obj, default=default, *args, **kwargs)
    
    @classmethod
    def _decode(cls, obj: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        return _simdjson.loads(obj, *args, **kwargs)
    
    @classmethod
    def parse(cls, obj: Any, *args, **kwargs) -> Union[Union[_simdjson.Object, _simdjson.Array], Union[Dict[Any, Any], List[str]]]:
        if not isinstance(obj, (str, bytes)):
            if getattr(obj, 'content', None): return cls.parser.parse(obj.content, *args, **kwargs)
        return cls.parser.parse(obj, *args, **kwargs)
    
    @classmethod
    def decode(cls, obj: Any, *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if isinstance(obj, (dict, list, set)): return obj
        if issubclass(obj, BaseModel): return obj.dict()
        if issubclass(obj, _simdjson.Object) or isinstance(obj, _simdjson.Object): return obj.as_dict()
        if issubclass(obj, _simdjson.Array) or isinstance(obj, _simdjson.Array): return obj.as_list()
        if isinstance(obj, (str, bytes)): return _simdjson.loads(obj)
        raise ValueError
    
    @classmethod
    async def _async_encode(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        return _simdjson.dumps(obj, default=default, *args, **kwargs)
    
    @classmethod
    async def _async_decode(cls, obj: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        return _simdjson.loads(obj, *args, **kwargs)



    


from __future__ import annotations

import simdjson as _simdjson
from typing import Dict, Union, Any, List, Optional, Type, Iterable, overload, Callable, Iterable, Iterator, TYPE_CHECKING
from typing_extensions import SupportsIndex
if TYPE_CHECKING:
    from _typeshed import SupportsLessThanT, SupportsLessThan

""" 
Extends the Object/Array C-types from pysimdjson
Makes them more python friendly, while keeping their performance benefits
"""

class SimdArray(object):
    def __init__(self, data: Union[_simdjson.Array, 'SimdArray', List[Any]]):
        self._obj = data
        self._data = []
    
    @property
    def data(self):
        if not self._data: 
            if not isinstance(self._obj, (_simdjson.Array, type(self))): 
                print(type(self._obj))
                return self._obj
            self._data = self._obj.as_list()
        return self._data

    def clear(self) -> None:
        self.data.clear()

    def copy(self) -> Type[List]:
        return self.data.copy()

    def append(self, __object: Any) -> None:
        self.data.append(__object)
    
    def extend(self, __iterable: Iterable[Any]) -> None:
        self.data.extend(__iterable)
    
    def pop(self, __index: SupportsIndex = ...) -> Any: 
        return self.data.pop(__index)

    def index(self, __value: Any, __start: SupportsIndex = ..., __stop: SupportsIndex = ...) -> int:
        return self.data.index(__value, __start, __stop)

    def count(self, __value: Any) -> int:
        return self.data.count(__value)

    def insert(self, __index: SupportsIndex, __object: Any) -> None:
        self.data.insert(__index, __object)

    def remove(self, __value: Any) -> None: 
        self.data.remove(__value)

    def reverse(self) -> None:
        self.data.reverse()

    @overload
    def sort(self, *, key: Callable[[Any], SupportsLessThan], reverse: bool = ...) -> None:
        self.data.sort(key, reverse)
    
    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[Any]:
        return self.data.__iter__

    def __str__(self) -> str:
        return self.data.__str__

    @overload
    def __getitem__(self, __i: SupportsIndex) -> Any:
        return self.data.__getitem__(__i)

    @overload
    def __getitem__(self, __s: slice) -> list[Any]:
        return self.data.__getitem__(__s)

    @overload
    def __setitem__(self, __i: SupportsIndex, __o: Any) -> None:
        self.data.__setitem__(__i, __o)

    @overload
    def __setitem__(self, __s: slice, __o: Iterable[Any]) -> None:
        self.data.__setitem__(__s, __o)

    def __delitem__(self, __i: Union[SupportsIndex, slice]) -> None:
        self.data.__delitem__(__i)

    def __add__(self, __x: list[Any]) -> list[Any]:
        return self.data.__add__(__x)

    def __iadd__(self, __x: Iterable[Any]) -> Any:
        return self.data.__iadd__(__x)

    def __mul__(self, __n: SupportsIndex) -> list[Any]:
        return self.data.__mul__(__n)

    def __rmul__(self, __n: SupportsIndex) -> list[Any]:
        return self.data.__rmul__(__n)

    def __imul__(self, __n: SupportsIndex) -> Any:
        return self.data.__imul__(__n)

    def __contains__(self, __o: object) -> bool:
        return self.data.__contains__(__o)

    def __reversed__(self) -> Iterator[Any]:
        return self.data.__reversed__()

    def __gt__(self, __x: list[Any]) -> bool:
        return self.data.__gt__(__x)

    def __ge__(self, __x: list[Any]) -> bool:
        return self.data.__ge__(__x)

    def __lt__(self, __x: list[Any]) -> bool:
        return self.data.__lt__(__x)

    def __le__(self, __x: list[Any]) -> bool:
        return self.data.__le__(__x)


class SimdObject(object):
    def __init__(self, data: Union[_simdjson.Object, 'SimdObject', Dict[Any, Any]]):
        self._obj = data
        self._data = {}

    @property
    def data(self) -> Dict[Any, Any]:
        if not self._data: 
            if not isinstance(self._obj, (_simdjson.Object, type(self))): 
                print(type(self._obj))
                return self._obj
            self._data = self._obj.as_dict()
        return self._data

    def dict(self) -> Dict[Any, Any]:
        return self.data

    @property
    def keys(self):
        return list(self.data.keys())

    @property
    def values(self):
        return list(self.data.values())
    
    @property
    def items(self):
        return dict(self.data.items())
    
    @property
    def json(self):
        return _simdjson.dumps(self.data)

    def update(self, **kwargs):
        self.data.update(**kwargs)

    def get(self, key: str, default: Any = None):
        if self._data.get(key): return self._data[key]
        return type(self)(self._obj.get(key, default))
    
    def pop(self, key: str, default: Any = None):
        return self.data.pop(key, default)
    
    def popitem(self): 
        return self.data.popitem()

    def copy(self):
        return self.data.copy()

    def __iter__(self):
        return self.data.__iter__

    def __delitem__(self, key: str):
        self.data.__delitem__(key)

    def __setitem__(self, key: str, value: Any = None):
        self.data.__setitem__(key, value)

    def __getitem__(self, key: str, default: Any = None) -> Any:
        if self._data.get(key): return self._data[key]
        d = type(self)(self._obj.get(key, default))
        return d.data

    def __len__(self):
        return self.data.__len__
    
    def __str__(self):
        return self.json

def create_simdobj(data: Union[_simdjson.Object, _simdjson.Array]):
    if isinstance(data, _simdjson.Object): return SimdObject(data)
    return SimdArray(data)


import string
import secrets
from typing import Dict
from .static import Defaults
from ._base import Base

ALPHA_NUMERIC = string.ascii_letters + string.digits

class Secret:
    
    @classmethod
    def uuid_passcode(cls, length: int = None, clean: bool = True, method: str = Defaults.uuid_method):
        rez = Base.get_uuid(method=method)
        if clean: rez = rez.replace('-', '').strip()
        if length: rez = rez[:length]
        return rez
    
    @classmethod
    def alphanumeric_passcode(cls, length: int = 16):
        return ''.join(secrets.choice(ALPHA_NUMERIC) for i in range(length))

    @classmethod
    def token(cls, length: int = 32, safe: bool = False, clean: bool = True):
        rez = secrets.token_hex(length) if safe else secrets.token_urlsafe(length)
        if clean:
            for i in rez: 
                if i not in ALPHA_NUMERIC: rez.replace(i, secrets.choice(ALPHA_NUMERIC))
        return rez

    @classmethod
    def keypair(cls, k_len: int = 16, s_len: int = 36) -> Dict[str, str]:
        return {
            'key': cls.alphanumeric_passcode(k_len),
            'secret': cls.alphanumeric_passcode(s_len) 
        }




from . import core
from . import cloud
from . import generic

from .core import run_fuze, BaseFuzerCls, MountPoint
from .cloud import GCSFuze, S3Fuze, MinioFuze, S3CompatFuze
from .generic import FuzeTypes, get_fuze, autofuze_mount, autofuze_unmount

__all__ = (
    'run_fuze',
    'BaseFuzerCls',
    'MountPoint',
    'GCSFuze',
    'S3Fuze',
    'MinioFuze',
    'S3CompatFuze',
    'FuzeTypes',
    'get_fuze',
    'autofuze_mount',
    'autofuze_unmount'
)


import os
import signal
import threading
import subprocess
import multiprocessing

from typing import Union
from types import ModuleType

try: import psutil
except: psutil: ModuleType = None


def get_mp(require_singlethread: bool = False):
    # We can't use forkserver because we have to make sure
    # that the server inherits the per-test stdout/stderr file
    # descriptors.
    if hasattr(multiprocessing, 'get_context'): mp = multiprocessing.get_context('fork')
    else: mp = multiprocessing
    if require_singlethread and threading.active_count() != 1: raise RuntimeError("Single-threaded is required")
    return mp

def exitcode(process: Union[subprocess.Popen, multiprocessing.Process]):
    if isinstance(process, subprocess.Popen): return process.poll()
    if process.is_alive(): return None
    else: return process.exitcode

def process_kill(pid: int, sig: int = signal.SIGTERM):
    """
    Used to kill a running process
    """
    os.kill(pid, sig)

def process_alive(pid: int) -> bool:
    """
    Used to check if a process is running
    """
    global psutil
    if not psutil:
        from lazy.libz import Lib
        psutil = Lib['psutil']
    return psutil.pid_exists(pid)


__all__ = (
    get_mp,
    exitcode,
    process_kill,
    process_alive
)